"""
Continuous variables have values anywhere in the range minus
infinite to plus infinite. However, when creating a visual
representation of these values what usually matters is the
relative difference between the values. This is where rescaling
comes into play.

The values are mapped onto a range that a scale can deal with. For
graphical representation that range tends to be :math:`[0, 1]` or
:math:`[0, n]`, where :math:`n` is some number that makes the
plotted object overflow the plotting area.

Although a scale may be able handle the :math:`[0, n]` range, it
may be desirable to have a lower bound greater than zero. For
example, if data values get mapped to zero on a scale whose
graphical representation is the size/area/radius/length some data
will be invisible. The solution is to restrict the lower bound
e.g. :math:`[0.1, 1]`. Similarly you can restrict the upper bound
-- using these functions.
"""

import datetime

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from .utils import first_element

__all__ = [
    "censor",
    "expand_range",
    "rescale",
    "rescale_max",
    "rescale_mid",
    "squish_infinite",
    "zero_range",
    "expand_range_distinct",
    "squish",
]


def rescale(x, to=(0, 1), _from=None):
    """
    Rescale numeric vector to have specified minimum and maximum.

    Parameters
    ----------
    x : array_like | numeric
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> x = [0, 2, 4, 6, 8, 10]
    >>> rescale(x)
    array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
    >>> rescale(x, to=(0, 2))
    array([0. , 0.4, 0.8, 1.2, 1.6, 2. ])
    >>> rescale(x, to=(0, 2), _from=(0, 20))
    array([0. , 0.2, 0.4, 0.6, 0.8, 1. ])
    """
    if _from is None:
        _from = np.min(x), np.max(x)
    return np.interp(x, _from, to)


def rescale_mid(x, to=(0, 1), _from=None, mid=0):
    """
    Rescale numeric vector to have specified minimum, midpoint,
    and maximum.

    Parameters
    ----------
    x : array_like | numeric
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x
    mid	: numeric
        mid-point of input range

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> rescale_mid([1, 2, 3], mid=1)
    array([0.5 , 0.75, 1.  ])
    >>> rescale_mid([1, 2, 3], mid=2)
    array([0. , 0.5, 1. ])
    """
    array_like = True

    try:
        len(x)
    except TypeError:
        array_like = False
        x = [x]

    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    if _from is None:
        _from = np.array([np.min(x), np.max(x)])
    else:
        _from = np.asarray(_from)

    if zero_range(_from) or zero_range(to):
        out = np.repeat(np.mean(to), len(x))
    else:
        extent = 2 * np.max(np.abs(_from - mid))
        out = (x - mid) / extent * np.diff(to) + np.mean(to)

    if not array_like:
        out = out[0]
    return out


def rescale_max(x, to=(0, 1), _from=None):
    """
    Rescale numeric vector to have specified maximum.

    Parameters
    ----------
    x : array_like | numeric
        1D vector of values to manipulate.
    to : tuple
        output range (numeric vector of length two)
    _from : tuple
        input range (numeric vector of length two).
        If not given, is calculated from the range of x.
        Only the 2nd (max) element is essential to the
        output.

    Returns
    -------
    out : array_like
        Rescaled values

    Examples
    --------
    >>> x = [0, 2, 4, 6, 8, 10]
    >>> rescale_max(x, (0, 3))
    array([0. , 0.6, 1.2, 1.8, 2.4, 3. ])

    Only the 2nd (max) element of the parameters ``to``
    and ``_from`` are essential to the output.

    >>> rescale_max(x, (1, 3))
    array([0. , 0.6, 1.2, 1.8, 2.4, 3. ])
    >>> rescale_max(x, (0, 20))
    array([ 0.,  4.,  8., 12., 16., 20.])

    If :python:`max(x) < _from[1]` then values will be
    scaled beyond the requested maximum (:python:`to[1]`).

    >>> rescale_max(x, to=(1, 3), _from=(-1, 6))
    array([0., 1., 2., 3., 4., 5.])

    If the values are the same, they taken on the requested maximum.
    This includes an array of all zeros.

    >>> rescale_max([5, 5, 5])
    array([1., 1., 1.])
    >>> rescale_max([0, 0, 0])
    array([1, 1, 1])
    """
    array_like = True

    try:
        len(x)
    except TypeError:
        array_like = False
        x = [x]

    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    if _from is None:
        _from = (np.min(x), np.max(x))

    if np.any(x < 0):
        out = rescale(x, (0, to[1]), _from)
    elif np.all(x == 0) and _from[1] == 0:
        out = np.repeat(to[1], len(x))
    else:
        out = x / _from[1] * to[1]

    if not array_like:
        out = out[0]
    return out


def squish_infinite(x, range=(0, 1)):
    """
    Truncate infinite values to a range.

    Parameters
    ----------
    x : array_like
        Values that should have infinities squished.
    range : tuple
        The range onto which to squish the infinites.
        Must be of size 2.

    Returns
    -------
    out : array_like
        Values with infinites squished.

    Examples
    --------
    >>> squish_infinite([0, .5, .25, np.inf, .44])
    [0.0, 0.5, 0.25, 1.0, 0.44]
    >>> squish_infinite([0, -np.inf, .5, .25, np.inf], (-10, 9))
    [0.0, -10.0, 0.5, 0.25, 9.0]
    """
    xtype = type(x)

    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    x[x == -np.inf] = range[0]
    x[x == np.inf] = range[1]

    if not isinstance(x, xtype):
        x = xtype(x)
    return x


def squish(x, range=(0, 1), only_finite=True):
    """
    Squish values into range.

    Parameters
    ----------
    x : array_like
        Values that should have out of range values squished.
    range : tuple
        The range onto which to squish the values.
    only_finite: boolean
        When true, only squishes finite values.

    Returns
    -------
    out : array_like
        Values with out of range values squished.

    Examples
    --------
    >>> squish([-1.5, 0.2, 0.5, 0.8, 1.0, 1.2])
    [0.0, 0.2, 0.5, 0.8, 1.0, 1.0]

    >>> squish([-np.inf, -1.5, 0.2, 0.5, 0.8, 1.0, np.inf], only_finite=False)
    [0.0, 0.0, 0.2, 0.5, 0.8, 1.0, 1.0]
    """
    xtype = type(x)

    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    finite = np.isfinite(x) if only_finite else True

    x[np.logical_and(x < range[0], finite)] = range[0]
    x[np.logical_and(x > range[1], finite)] = range[1]

    if not isinstance(x, xtype):
        x = xtype(x)
    return x


def censor(x, range=(0, 1), only_finite=True):
    """
    Convert any values outside of range to a **NULL** type object.

    Parameters
    ----------
    x : array_like
        Values to manipulate
    range : tuple
        (min, max) giving desired output range
    only_finite : bool
        If True (the default), will only modify
        finite values.

    Returns
    -------
    x : array_like
        Censored array

    Examples
    --------
    >>> a = [1, 2, np.inf, 3, 4, -np.inf, 5]
    >>> censor(a, (0, 10))
    [1, 2, inf, 3, 4, -inf, 5]
    >>> censor(a, (0, 10), False)
    [1, 2, nan, 3, 4, nan, 5]
    >>> censor(a, (2, 4))
    [nan, 2, inf, 3, 4, -inf, nan]

    Notes
    -----
    All values in ``x`` should be of the same type. ``only_finite`` parameter
    is not considered for Datetime and Timedelta types.

    The **NULL** type object depends on the type of values in **x**.

    - :class:`float` - :py:`float('nan')`
    - :class:`int` - :py:`float('nan')`
    - :class:`datetime.datetime` : :py:`np.datetime64(NaT)`
    - :class:`datetime.timedelta` : :py:`np.timedelta64(NaT)`

    """
    if not len(x):
        return x

    py_time_types = (datetime.datetime, datetime.timedelta)
    np_pd_time_types = (
        pd.Timestamp,
        pd.Timedelta,
        np.datetime64,
        np.timedelta64,
    )
    x0 = first_element(x)

    # Yes, we want type not isinstance
    if type(x0) in py_time_types:
        return _censor_with(x, range, "NaT")

    if not hasattr(x, "dtype") and isinstance(x0, np_pd_time_types):
        return _censor_with(x, range, type(x0)("NaT"))

    x_array = np.asarray(x)
    if pdtypes.is_number(x0) and not isinstance(x0, np.timedelta64):
        null = float("nan")
    elif isinstance(x0, pd.Timestamp):
        null = pd.Timestamp("NaT")
    elif pdtypes.is_datetime64_dtype(x_array):
        null = np.datetime64("NaT")
    elif isinstance(x0, pd.Timedelta):
        null = pd.Timedelta("NaT")
    elif pdtypes.is_timedelta64_dtype(x_array):
        null = np.timedelta64("NaT")
    else:
        raise ValueError(
            "Do not know how to censor values of type " "{}".format(type(x0))
        )

    if only_finite:
        try:
            finite = np.isfinite(x)
        except TypeError:
            finite = np.repeat(True, len(x))
    else:
        finite = np.repeat(True, len(x))

    if hasattr(x, "dtype"):
        # Ignore RuntimeWarning when x contains nans
        with np.errstate(invalid="ignore"):
            outside = (x < range[0]) | (x > range[1])
        bool_idx = finite & outside
        x = x.copy()
        x[bool_idx] = null
    else:
        x = [
            null if not range[0] <= val <= range[1] and f else val
            for val, f in zip(x, finite)
        ]

    return x


def _censor_with(x, range, value=None):
    """
    Censor any values outside of range with ``None``
    """
    return [val if range[0] <= val <= range[1] else value for val in x]


def zero_range(x, tol=np.finfo(float).eps * 100):
    """
    Determine if range of vector is close to zero.

    Parameters
    ----------
    x : array_like | numeric
        Value(s) to check. If it is an array_like, it
        should be of length 2.
    tol : float
        Tolerance. Default tolerance is the `machine epsilon`_
        times :math:`10^2`.

    Returns
    -------
    out : bool
        Whether ``x`` has zero range.

    Examples
    --------
    >>> zero_range([1, 1])
    True
    >>> zero_range([1, 2])
    False
    >>> zero_range([1, 2], tol=2)
    True

    .. _machine epsilon: https://en.wikipedia.org/wiki/Machine_epsilon
    """
    try:
        if len(x) == 1:
            return True
    except TypeError:
        return True

    if len(x) != 2:
        raise ValueError("x must be length 1 or 2")

    # Also deals with array_likes that have non-standard indices
    x = sorted(x)
    low, high = x

    # datetime - pandas, cpython
    if isinstance(low, (pd.Timestamp, datetime.datetime)):
        from matplotlib.dates import date2num

        # date2num include timezone info, .toordinal() does not
        low, high = date2num(x)
    # datetime - numpy
    elif isinstance(low, np.datetime64):
        return low == high
    # timedelta - pandas, cpython
    elif isinstance(low, (pd.Timedelta, datetime.timedelta)):
        low, high = low.total_seconds(), high.total_seconds()
    # timedelta - numpy
    elif isinstance(low, np.timedelta64):
        return low == high
    elif not isinstance(low, (float, int, np.number)):
        raise TypeError(
            "zero_range objects cannot work with objects "
            "of type '{}'".format(type(low))
        )

    if any(np.isnan((low, high))):
        return np.nan

    if low == high:
        return True

    if any(np.isinf((low, high))):
        return False

    low_abs = np.abs(low)
    if low_abs == 0:
        return False

    return ((high - low) / low_abs) < tol


def expand_range(range, mul=0, add=0, zero_width=1):
    """
    Expand a range with a multiplicative or additive constant

    Parameters
    ----------
    range : tuple
        Range of data. Size 2.
    mul : int | float
        Multiplicative constant
    add : int | float | timedelta
        Additive constant
    zero_width : int | float | timedelta
        Distance to use if range has zero width

    Returns
    -------
    out : tuple
        Expanded range

    Examples
    --------
    >>> expand_range((3, 8))
    (3, 8)
    >>> expand_range((0, 10), mul=0.1)
    (-1.0, 11.0)
    >>> expand_range((0, 10), add=2)
    (-2, 12)
    >>> expand_range((0, 10), mul=.1, add=2)
    (-3.0, 13.0)
    >>> expand_range((0, 1))
    (0, 1)

    When the range has zero width

    >>> expand_range((5, 5))
    (4.5, 5.5)

    Notes
    -----
    If expanding *datetime* or *timedelta* types, **add** and
    **zero_width** must be suitable *timedeltas* i.e. You should
    not mix types between **Numpy**, **Pandas** and the
    :mod:`datetime` module.
    """
    x = range

    # Enforce tuple
    try:
        x[0]
    except TypeError:
        x = (x, x)

    # The expansion cases
    if zero_range(x):
        new = x[0] - zero_width / 2, x[0] + zero_width / 2
    else:
        dx = (x[1] - x[0]) * mul + add
        new = x[0] - dx, x[1] + dx

    return new


def expand_range_distinct(range, expand=(0, 0, 0, 0), zero_width=1):
    """
    Expand a range with a multiplicative or additive constants

    Similar to :func:`expand_range` but both sides of the range
    expanded using different constants

    Parameters
    ----------
    range : tuple
        Range of data. Size 2
    expand : tuple
        Length 2 or 4. If length is 2, then the same constants
        are used for both sides. If length is 4 then the first
        two are are the Multiplicative (*mul*) and Additive (*add*)
        constants for the lower limit, and the second two are
        the constants for the upper limit.
    zero_width : int | float | timedelta
        Distance to use if range has zero width

    Returns
    -------
    out : tuple
        Expanded range

    Examples
    --------
    >>> expand_range_distinct((3, 8))
    (3, 8)
    >>> expand_range_distinct((0, 10), (0.1, 0))
    (-1.0, 11.0)
    >>> expand_range_distinct((0, 10), (0.1, 0, 0.1, 0))
    (-1.0, 11.0)
    >>> expand_range_distinct((0, 10), (0.1, 0, 0, 0))
    (-1.0, 10)
    >>> expand_range_distinct((0, 10), (0, 2))
    (-2, 12)
    >>> expand_range_distinct((0, 10), (0, 2, 0, 2))
    (-2, 12)
    >>> expand_range_distinct((0, 10), (0, 0, 0, 2))
    (0, 12)
    >>> expand_range_distinct((0, 10), (.1, 2))
    (-3.0, 13.0)
    >>> expand_range_distinct((0, 10), (.1, 2, .1, 2))
    (-3.0, 13.0)
    >>> expand_range_distinct((0, 10), (0, 0, .1, 2))
    (0, 13.0)
    """

    if len(expand) == 2:
        expand = tuple(expand) * 2

    lower = expand_range(range, expand[0], expand[1], zero_width)[0]
    upper = expand_range(range, expand[2], expand[3], zero_width)[1]
    return (lower, upper)



# -*- coding: utf-8 -*-
import warnings
from datetime import datetime, timedelta, tzinfo

import numpy as np
import pandas as pd
import pytest
import pytz

try:
    from zoneinfo import ZoneInfo
except ImportError:
    # python < 3.9
    from backports.zoneinfo import ZoneInfo

from mizani.formatters import (
    comma_format,
    currency_format,
    custom_format,
    date_format,
    log_format,
    mpl_format,
    number_bytes_format,
    ordinal_format,
    percent_format,
    pvalue_format,
    scientific_format,
    timedelta_format,
)


def test_custom_format():
    x = [3.987, 2, 42.42]
    labels = ["3.99 USD", "2.00 USD", "42.42 USD"]
    formatter = custom_format("{:.2f} USD")
    assert formatter(x) == labels

    formatter = custom_format("%.2f USD", style="old")
    assert formatter(x) == labels

    formatter = custom_format("%.2f USD", style="ancient")
    with pytest.raises(ValueError):
        formatter(x)


def test_currency_format():
    x = [1.232, 99.2334, 4.6, 9, 4500]
    formatter = currency_format("C$", digits=0, big_mark=",")
    result = formatter(x)
    assert result == ["C$1", "C$99", "C$5", "C$9", "C$4,500"]

    formatter = currency_format("C$", digits=0, big_mark=" ")
    result = formatter(x)
    assert result == ["C$1", "C$99", "C$5", "C$9", "C$4 500"]

    formatter = currency_format("$", digits=2)
    result = formatter(x)
    assert result == ["$1.23", "$99.23", "$4.60", "$9.00", "$4500.00"]


def test_comma_format():
    x = [1000, 2, 33000, 400]
    result = comma_format()(x)
    assert result == ["1,000", "2", "33,000", "400"]


def test_percent_format():
    formatter = percent_format()
    # same/nearly same precision values
    assert formatter([0.12, 0.23, 0.34, 0.45]) == ["12%", "23%", "34%", "45%"]

    assert formatter([0.12, 0.23, 0.34, 4.5]) == ["12%", "23%", "34%", "450%"]

    # mixed precision values
    assert formatter([0.12, 0.23, 0.34, 45]) == ["10%", "20%", "30%", "4500%"]


def test_scientific():
    formatter = scientific_format(2)
    assert formatter([0.12, 0.2376, 0.34, 45]) == [
        "1.20e-01",
        "2.38e-01",
        "3.40e-01",
        "4.50e+01",
    ]

    assert formatter([0.12, 230, 0.34 * 10**5, 0.4]) == [
        "1.2e-01",
        "2.3e+02",
        "3.4e+04",
        "4.0e-01",
    ]


def test_mpl_format():
    formatter = mpl_format()
    assert formatter([5, 10, 100, 150]) == ["5", "10", "100", "150"]

    # trigger the order of magnitude correction (not any more)
    assert formatter([5, 10, 100, 150e8]) == ["5", "10", "100", "15000000000"]


def test_log_format():
    formatter = log_format()
    assert formatter([0.001, 0.1, 100]) == ["0.001", "0.1", "100"]
    assert formatter([0.001, 0.1, 10000]) == ["1e-3", "1e-1", "1e4"]
    assert formatter([35, 60]) == ["35", "60"]
    assert formatter([34.99999999999, 60.0000000001]) == ["35", "60"]
    assert formatter([3000.0000000000014, 4999.999999999999]) == [
        "3000",
        "5000",
    ]
    assert formatter([1, 35, 60, 1000]) == ["1", "35", "60", "1000"]
    assert formatter([1, 35, 60, 10000]) == ["1", "35", "60", "10000"]
    assert formatter([3.000000000000001e-05]) == ["3e-5"]
    assert formatter([1, 1e4]) == ["1", "1e4"]
    assert formatter([1, 35, 60, 1e6]) == ["1", "4e1", "6e1", "1e6"]

    formatter = log_format(base=2)
    assert formatter([1, 2, 4, 8]) == ["2^0", "2^1", "2^2", "2^3"]
    assert formatter([0b1, 0b10, 0b11]) == ["2^0", "2^1", "2^1.585"]

    formatter = log_format(base=8)
    assert formatter([1, 4, 8, 64]) == ["8^0", "8^0.667", "8^1", "8^2"]

    formatter = log_format(base=5)
    assert formatter([1, 5, 25, 125]) == ["5^0", "5^1", "5^2", "5^3"]

    formatter = log_format(base=np.e)
    assert formatter([1, np.pi, np.e**2, np.e**3]) == [
        "e^0",
        "e^1.145",
        "e^2",
        "e^3",
    ]

    # mathtex
    formatter = log_format(mathtex=True)
    assert formatter([0.001, 0.1, 10000]) == [
        "$10^{-3}$",
        "$10^{-1}$",
        "$10^{4}$",
    ]
    assert formatter([35, 60]) == ["35", "60"]
    assert formatter([1, 10000]) == ["$10^{0}$", "$10^{4}$"]

    formatter = log_format(base=8, mathtex=True)
    assert formatter([1, 4, 64]) == ["$8^{0}$", "$8^{0.667}$", "$8^{2}$"]


def test_date_format():
    x = pd.date_range("1/1/2010", periods=4, freq="4AS")
    result = date_format("%Y")(x)
    assert result == ["2010", "2014", "2018", "2022"]

    x = [datetime(year=2005 + i, month=i, day=i) for i in range(1, 5)]
    result = date_format("%Y:%m:%d")(x)
    assert result == ["2006:01:01", "2007:02:02", "2008:03:03", "2009:04:04"]

    # Different timezones
    PCT = ZoneInfo("US/Pacific")
    UG = ZoneInfo("Africa/Kampala")
    x = [datetime(2010, 1, 1, tzinfo=UG), datetime(2010, 1, 1, tzinfo=PCT)]
    with pytest.warns(UserWarning, match=r"different time zones"):
        date_format()(x)

    # Timezone with Daylight time
    NY = ZoneInfo("America/New_York")
    x = [datetime(2023, 10, 1, tzinfo=NY), datetime(2023, 11, 1, tzinfo=NY)]
    result = date_format()(x)
    assert result == ["2023-10-01", "2023-11-01"]

    # Same as above, but different tz library
    NY = pytz.timezone("America/New_York")
    x = [datetime(2023, 10, 1, tzinfo=NY), datetime(2023, 11, 1, tzinfo=NY)]
    result = date_format()(x)
    assert result == ["2023-10-01", "2023-11-01"]

    # Unknown Timezone
    class myTzInfo(tzinfo):
        def __str__(self):
            return "None"

    TZ = myTzInfo()
    x = [datetime(2023, 10, 1, tzinfo=TZ), datetime(2023, 11, 1, tzinfo=TZ)]
    with pytest.raises(NotImplementedError, match=r"^a tzinfo subclass"):
        date_format()(x)


def test_timedelta_format():
    x = [timedelta(days=7 * i) for i in range(5)]
    labels = timedelta_format()(x)
    assert labels == ["0", "1 week", "2 weeks", "3 weeks", "4 weeks"]

    x = [pd.Timedelta(seconds=600 * i) for i in range(5)]
    labels = timedelta_format()(x)
    assert labels == [
        "0",
        "10 minutes",
        "20 minutes",
        "30 minutes",
        "40 minutes",
    ]

    # specific units
    labels = timedelta_format(units="h")(x)
    assert labels == [
        "0",
        "0.1667 hours",
        "0.3333 hours",
        "0.5000 hours",
        "0.6667 hours",
    ]
    # usetex
    x = [timedelta(microseconds=7 * i) for i in range(5)]
    labels = timedelta_format(units="us", usetex=True)(x)
    assert labels == [
        "0",
        "7$\\mu s$",
        "14$\\mu s$",
        "21$\\mu s$",
        "28$\\mu s$",
    ]


def test_pvalue_format():
    x = [0.90, 0.15, 0.015, 0.009, 0.0005]
    labels = pvalue_format()(x)
    assert labels == ["0.9", "0.15", "0.015", "0.009", "<0.001"]

    labels = pvalue_format(add_p=True)(x)
    assert labels == ["p=0.9", "p=0.15", "p=0.015", "p=0.009", "p<0.001"]

    with warnings.catch_warnings(record=True) as record:
        x = [0.90, 0.15, np.nan, 0.015, 0.009, 0.0005]
        labels = pvalue_format()(x)
        assert labels == ["0.9", "0.15", "nan", "0.015", "0.009", "<0.001"]
        assert not record, "Issued an unexpected warning"

    # NaN is handled without any warning
    assert len(record) == 0


def test_ordinal_format():
    labels = ordinal_format()(range(110, 115))
    assert labels == ["110th", "111th", "112th", "113th", "114th"]

    labels = ordinal_format()(range(120, 125))
    assert labels == ["120th", "121st", "122nd", "123rd", "124th"]

    labels = ordinal_format(big_mark=",")(range(1200, 1205))
    assert labels == ["1,200th", "1,201st", "1,202nd", "1,203rd", "1,204th"]

    labels = ordinal_format(big_mark=".")(range(1200, 1205))
    assert labels == ["1.200th", "1.201st", "1.202nd", "1.203rd", "1.204th"]


def test_number_bytes_format():
    x = [1000, 1000000, 4e5]
    labels = number_bytes_format(symbol="MiB")(x)
    assert labels == ["0 MiB", "1 MiB", "0 MiB"]

    labels = number_bytes_format(symbol="MiB", fmt="{:.2f} ")(x)
    assert labels == ["0.00 MiB", "0.95 MiB", "0.38 MiB"]

    with pytest.raises(ValueError):
        number_bytes_format(symbol="Bad")(x)


def test_empty_breaks():
    x = []
    assert custom_format()(x) == []
    assert comma_format()(x) == []
    assert currency_format()(x) == []
    assert percent_format()(x) == []
    assert scientific_format()(x) == []
    assert date_format()(x) == []
    assert mpl_format()(x) == []
    assert log_format()(x) == []
    assert timedelta_format()(x) == []
    assert pvalue_format()(x) == []
    assert ordinal_format()(x) == []
    assert number_bytes_format()(x) == []



from datetime import datetime, timedelta
from types import FunctionType, MethodType

try:
    from zoneinfo import ZoneInfo
except ImportError:
    # python < 3.9
    from backports.zoneinfo import ZoneInfo

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from mizani.breaks import minor_breaks, mpl_breaks
from mizani.transforms import (
    asn_trans,
    atanh_trans,
    boxcox_trans,
    datetime_trans,
    exp_trans,
    gettrans,
    identity_trans,
    log1p_trans,
    log2_trans,
    log10_trans,
    log_trans,
    modulus_trans,
    pd_timedelta_trans,
    probability_trans,
    pseudo_log_trans,
    reciprocal_trans,
    reverse_trans,
    sqrt_trans,
    timedelta_trans,
    trans,
    trans_new,
)

arr = np.arange(1, 100)


def test_trans():
    with pytest.raises(KeyError):
        trans(universe=True)


def test_trans_new():
    t = trans_new(
        "bounded_identity",
        staticmethod(lambda x: x),
        classmethod(lambda x: x),
        _format=lambda x: str(x),
        domain=(-999, 999),
        doc="Bounded Identity transform",
    )

    assert t.__name__ == "bounded_identity_trans"
    assert isinstance(t.transform, FunctionType)
    assert isinstance(t.inverse, MethodType)
    assert isinstance(t.format, FunctionType)
    assert t.domain == (-999, 999)
    assert t.__doc__ == "Bounded Identity transform"

    # ticks do not go beyond the bounds
    major = t().breaks((-1999, 1999))
    assert min(major) >= -999
    assert max(major) <= 999


def test_gettrans():
    t0 = identity_trans()
    t1 = gettrans(t0)
    t2 = gettrans(identity_trans)
    t3 = gettrans("identity")
    assert all(isinstance(x, identity_trans) for x in (t0, t1, t2, t3))

    t = gettrans(exp_trans)
    assert t.__class__.__name__ == "power_e_trans"

    with pytest.raises(ValueError):
        gettrans(object)


def _test_trans(trans, x):
    t = gettrans(trans())
    xt = t.transform(x)
    x2 = t.inverse(xt)
    is_log_trans = "log" in t.__class__.__name__ and hasattr(t, "base")
    # round trip
    npt.assert_allclose(x, x2)
    major = t.breaks([min(x), max(x)])
    minor = t.minor_breaks(t.transform(major))
    # Breaks and they are finite
    assert len(major)
    if is_log_trans and int(t.base) == 2:
        # Minor breaks for base == 2
        assert len(minor) == 0
    else:
        assert len(minor)
    assert all(np.isfinite(major))
    assert all(np.isfinite(minor))
    # Not breaks outside the domain
    assert all(major >= t.domain[0])
    assert all(major <= t.domain[1])
    assert all(minor >= t.domain[0])
    assert all(minor <= t.domain[1])


def test_asn_trans():
    _test_trans(asn_trans, arr * 0.01),


def test_atanh_trans():
    _test_trans(atanh_trans, arr * 0.001),


def test_boxcox_trans():
    _test_trans(boxcox_trans(0.5), arr * 10)
    _test_trans(boxcox_trans(1), arr)
    with pytest.raises(ValueError):
        x = np.arange(-4, 4)
        _test_trans(boxcox_trans(0.5), x)

    # Special case, small p and p = 0
    with pytest.warns(RuntimeWarning):
        _test_trans(boxcox_trans(1e-8), arr)
        _test_trans(boxcox_trans(0), arr)

    x = [0, 1, 2, 3]
    t = boxcox_trans(0)
    with pytest.warns(RuntimeWarning):
        xt = t.transform(x)
    xti = t.inverse(xt)
    assert np.isneginf(xt[0])
    npt.assert_array_almost_equal(x, xti)


def test_modulus_trans():
    _test_trans(modulus_trans(0), arr)
    _test_trans(modulus_trans(0.5), arr * 10)


def test_exp_trans():
    _test_trans(exp_trans, arr)

    exp2_trans = exp_trans(2)
    _test_trans(exp2_trans, arr * 0.1)


def test_identity_trans():
    _test_trans(identity_trans, arr)


def test_log10_trans():
    _test_trans(log10_trans, arr)


def test_log1p_trans():
    _test_trans(log1p_trans, arr)


def test_log2_trans():
    _test_trans(log2_trans, arr)


def test_log_trans():
    _test_trans(log_trans, arr)


def test_reverse_trans():
    _test_trans(reverse_trans, arr)


def test_sqrt_trans():
    _test_trans(sqrt_trans, arr)


def test_logn_trans():
    log3_trans = log_trans(3)
    _test_trans(log3_trans, arr)

    log4_trans = log_trans(
        4, domain=(0.1, 100), breaks=mpl_breaks(), minor_breaks=minor_breaks()
    )
    _test_trans(log4_trans, arr)


def test_reciprocal_trans():
    x = np.arange(10, 21)
    _test_trans(reciprocal_trans, x)


def test_pseudo_log_trans():
    p = np.arange(-4, 4)
    pos = [10 ** int(x) for x in p]
    arr = np.hstack([-np.array(pos[::-1]), pos])
    _test_trans(pseudo_log_trans, arr)
    _test_trans(pseudo_log_trans(base=16), arr)
    _test_trans(pseudo_log_trans(base=10, minor_breaks=minor_breaks(n=5)), arr)


def test_probability_trans():
    with pytest.raises(ValueError):
        t = probability_trans("unknown_distribution")

    # cdf of the normal is centered at 0 and
    # The values either end of 0 are symmetric
    x = [-3, -2, -1, 0, 1, 2, 3]
    t = probability_trans("norm")
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert xt[3] == 0.5
    npt.assert_allclose(xt[:3], 1 - xt[-3:][::-1])
    npt.assert_allclose(x, x2)


def test_datetime_trans():
    UTC = ZoneInfo("UTC")

    x = [datetime(year, 1, 1, tzinfo=UTC) for year in [2010, 2015, 2020, 2026]]
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))

    # numpy datetime64
    x = [np.datetime64(i, "D") for i in range(1, 11)]
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(isinstance(val, datetime) for val in x2)

    # pandas timestamp
    x = pd.date_range(start="1/1/2022", end="1/2/2022", freq="3H", tz="EST")
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(x == x2)

    # Irregular index
    s = pd.Series(x, index=range(11, len(x) + 11))
    st = t.transform(s)
    s2 = t.inverse(st)
    assert all(s == s2)

    # Scalar
    x = datetime(2022, 1, 20, tzinfo=UTC)
    t = datetime_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert x == x2


def test_datetime_trans_tz():
    EST = ZoneInfo("EST")
    UTC = ZoneInfo("UTC")

    x = [datetime(2022, 1, 1, 0 + 3 * i, 0, 0, tzinfo=EST) for i in range(8)]

    # Same trans as data
    t = datetime_trans()
    x2 = t.inverse(t.transform(x))
    assert x == x2
    assert all(val.tzinfo == EST for val in x2)

    # UTC trans
    t = datetime_trans(UTC)
    x2 = t.inverse(t.transform(x))
    assert x == x2
    assert all(val.tzinfo == UTC for val in x2)

    t = datetime_trans("MST")
    assert t.tzinfo == t.tz
    assert t.transform([]) == []


def test_timedelta_trans():
    x = [timedelta(days=i) for i in range(1, 11)]
    t = timedelta_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))
    assert x[0] == t.inverse(t.transform(x[0]))


def test_pd_timedelta_trans():
    x = [pd.Timedelta(days=i) for i in range(1, 11)]
    t = pd_timedelta_trans()
    xt = t.transform(x)
    x2 = t.inverse(xt)
    assert all(a == b for a, b in zip(x, x2))
    assert x[0] == t.inverse(t.transform(x[0]))



import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from mizani.bounds import rescale
from mizani.scale import scale_continuous, scale_discrete
from mizani.transforms import identity_trans


def test_scale_continuous():
    x = np.arange(11)
    # apply
    scaled = scale_continuous.apply(x, rescale)
    npt.assert_allclose(scaled, x * 0.1)
    npt.assert_allclose(scaled, x * 0.1)

    # Train
    limits = scale_continuous.train(x)
    npt.assert_allclose(limits, [0, 10])
    # Additional training
    limits = scale_continuous.train(np.arange(-4, 11), limits)
    npt.assert_allclose(limits, [-4, 10])
    limits = scale_continuous.train([], limits)
    npt.assert_allclose(limits, [-4, 10])

    # branches #
    scaled = scale_continuous.apply(x, rescale, trans=identity_trans())
    mapped = scale_continuous.map(x, list, [0, 10])
    npt.assert_allclose(mapped, x * 0.1)

    with pytest.raises(TypeError):
        # discrete data
        limits = scale_continuous.train(["a", "b", "c"])


def test_scale_discrete():
    def assert_equal_with_nan(lst1, lst2):
        assert lst1[:-1] == lst2[:-1] and np.isnan(lst2[-1])

    def SCategorical(*args, **kwargs):
        return pd.Series(pd.Categorical(*args, **kwargs))

    x = ["a", "b", "c", "a"]
    # apply
    scaled = scale_discrete.apply(x, np.arange)
    npt.assert_allclose(scaled, [0, 1, 2, 0])

    # Train
    limits = scale_discrete.train(x)
    assert limits == ["a", "b", "c"]

    # Additional training
    limits = scale_discrete.train(["b", "c", "d"], limits)
    assert limits == ["a", "b", "c", "d"]
    limits = scale_discrete.train([], limits)
    assert limits == ["a", "b", "c", "d"]
    limits = scale_discrete.train([None, "e"], limits)
    assert_equal_with_nan(limits, ["a", "b", "c", "d", "e", np.nan])

    # Train (Deal with missing values)
    x1 = ["a", "b", np.nan, "c"]
    x2 = pd.Categorical([1, 2, 3, np.nan])
    limits = scale_discrete.train(x1, na_rm=True)
    assert limits == ["a", "b", "c"]

    limits = scale_discrete.train(x1, na_rm=False)
    assert_equal_with_nan(limits, ["a", "b", "c", np.nan])

    limits = scale_discrete.train(x2, na_rm=True)
    assert limits == [1, 2, 3]

    limits = scale_discrete.train(x2, na_rm=False)
    assert_equal_with_nan(limits, [1, 2, 3, np.nan])

    # Double trainning with nan does not add another nan
    limits2 = scale_discrete.train(x2, limits, na_rm=False)
    assert_equal_with_nan(limits2, [1, 2, 3, np.nan])

    # branches #

    with pytest.raises(TypeError):
        # continuous data
        limits = scale_discrete.train([1, 2, 3])

    x = pd.Categorical(["a", "b"])
    limits = scale_discrete.train(x)
    assert limits == ["a", "b"]

    x = pd.Series(["a", "b", "c"], dtype="category")
    limits = scale_discrete.train(x[:1], drop=True)
    assert limits == ["a"]

    limits = scale_discrete.train(x[:2], drop=False)
    assert limits == ["a", "b", "c"]

    # Disrete Scale training maintains order of categoricals
    cats = ["x0", "x1", "x2", "x3", "x4"]
    s1 = SCategorical(["x1", "x2"], categories=cats)
    s2 = SCategorical(["x0", "x2"], categories=cats)
    limits = scale_discrete.train(s1, drop=True)
    limits = scale_discrete.train(s2, limits, drop=True)
    assert limits == ["x0", "x1", "x2"]

    # Trainning on mixed categories, the last data determines
    # the location of a value that is in two categoricals
    # eg. a & e are ordered right!
    x1 = pd.Categorical(["a", "b", "c", "e"])
    x2 = pd.Categorical(["d", "f", "e", "a"])
    limits = scale_discrete.train(x1)
    limits = scale_discrete.train(x2, old=limits)
    # assert limits == list('abcedf')
    assert limits == list("adefbc")



from datetime import datetime, timedelta

import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest

from mizani.breaks import (
    date_breaks,
    extended_breaks,
    log_breaks,
    minor_breaks,
    mpl_breaks,
    timedelta_breaks,
    trans_minor_breaks,
)
from mizani.transforms import log_trans, trans


def test_mpl_breaks():
    x = np.arange(100)
    limits = min(x), max(x)
    for nbins in (5, 7, 10, 13, 31):
        breaks = mpl_breaks(nbins=nbins)
        assert len(breaks(limits)) <= nbins + 1

    limits = float("-inf"), float("inf")
    breaks = mpl_breaks(nbins=5)
    assert len(breaks(limits)) == 0

    # Zero range discrete
    limits = [1, 1]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]

    # Zero range continuous
    limits = [np.pi, np.pi]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]


def test_log_breaks():
    x = [2, 20, 2000]
    limits = min(x), max(x)
    breaks = log_breaks()(limits)
    npt.assert_array_equal(breaks, [1, 10, 100, 1000, 10000])

    breaks = log_breaks(3)(limits)
    npt.assert_array_equal(breaks, [1, 100, 10000])

    breaks = log_breaks()((10000, 10000))
    npt.assert_array_equal(breaks, [10000])

    breaks = log_breaks()((float("-inf"), float("inf")))
    assert len(breaks) == 0

    # When the limits are in the same order of magnitude
    breaks = log_breaks()([35, 60])
    assert len(breaks) > 0
    assert all(1 < b < 100 for b in breaks)

    breaks = log_breaks()([200, 800])
    npt.assert_array_equal(breaks, [100, 200, 300, 500, 1000])

    breaks = log_breaks()((1664, 14008))
    npt.assert_array_equal(breaks, [1000, 3000, 5000, 10000, 30000])

    breaks = log_breaks()([407, 3430])
    npt.assert_array_equal(breaks, [300, 500, 1000, 3000, 5000])

    breaks = log_breaks()([1761, 8557])
    npt.assert_array_equal(breaks, [1000, 2000, 3000, 5000, 10000])

    # log_breaks -> _log_sub_breaks -> extended_breaks
    breaks = log_breaks(13)([1, 10])
    npt.assert_array_almost_equal(breaks, np.arange(0, 11))

    # No overflow effects
    breaks = log_breaks(n=6)([1e25, 1e30])
    npt.assert_array_almost_equal(breaks, [1e25, 1e26, 1e27, 1e28, 1e29, 1e30])

    # No overflow effects in _log_sub_breaks
    breaks = log_breaks()([2e19, 8e19])
    npt.assert_array_almost_equal(
        breaks, [1.0e19, 2.0e19, 3.0e19, 5.0e19, 1.0e20]
    )

    # _log_sub_breaks for base != 10
    breaks = log_breaks(n=5, base=60)([2e5, 8e5])
    npt.assert_array_almost_equal(
        breaks, [129600, 216000, 432000, 648000, 1080000]
    )

    breaks = log_breaks(n=5, base=2)([20, 80])
    npt.assert_array_almost_equal(breaks, [16, 32, 64, 128])

    # bases & negative breaks
    breaks = log_breaks(base=2)([0.9, 2.9])
    npt.assert_array_almost_equal(breaks, [0.5, 1.0, 2.0, 4.0])


def test_minor_breaks():
    # equidistant breaks
    major = [1, 2, 3, 4]
    limits = [0, 5]
    breaks = minor_breaks()(major, limits)
    npt.assert_array_equal(breaks, [0.5, 1.5, 2.5, 3.5, 4.5])
    minor = minor_breaks(3)(major, [2, 3])
    npt.assert_array_equal(minor, [2.25, 2.5, 2.75])

    # More than 1 minor breaks
    breaks = minor_breaks()(major, limits, 3)
    npt.assert_array_equal(
        breaks,
        [
            0.25,
            0.5,
            0.75,
            1.25,
            1.5,
            1.75,
            2.25,
            2.5,
            2.75,
            3.25,
            3.5,
            3.75,
            4.25,
            4.5,
            4.75,
        ],
    )

    # non-equidistant breaks
    major = [1, 2, 4, 8]
    limits = [0, 10]
    minor = minor_breaks()(major, limits)
    npt.assert_array_equal(minor, [1.5, 3, 6])

    # single major break
    minor = minor_breaks()([2], limits)
    assert len(minor) == 0


def test_trans_minor_breaks():
    class identity_trans(trans):
        def __init__(self):
            self.minor_breaks = trans_minor_breaks(identity_trans)

    class square_trans(trans):
        transform = staticmethod(np.square)
        inverse = staticmethod(np.sqrt)

        def __init__(self):
            self.minor_breaks = trans_minor_breaks(square_trans)

    class weird_trans(trans):
        dataspace_is_numerical = False

        def __init__(self):
            self.minor_breaks = trans_minor_breaks(weird_trans)

    major = [1, 2, 3, 4]
    limits = [0, 5]
    regular_minors = trans().minor_breaks(major, limits)
    npt.assert_allclose(
        regular_minors, identity_trans().minor_breaks(major, limits)
    )

    # Transform the input major breaks and check against
    # the inverse of the output minor breaks
    squared_input_minors = square_trans().minor_breaks(
        np.square(major), np.square(limits)
    )
    npt.assert_allclose(regular_minors, np.sqrt(squared_input_minors))

    t = weird_trans()
    with pytest.raises(TypeError):
        t.minor_breaks(major)

    # Test minor_breaks for log scales are 2 less than the base
    base = 10
    breaks = np.arange(1, 3)
    limits = [breaks[0], breaks[-1]]
    t = log_trans(base)
    assert len(t.minor_breaks(breaks, limits)) == base - 2

    base = 5  # Odd base
    breaks = np.arange(1, 3)
    limits = [breaks[0], breaks[-1]]
    t = log_trans(base)
    assert len(t.minor_breaks(breaks, limits)) == base - 2

    t = log_trans()
    major = t.transform([1, 10, 100])
    limits = t.transform([1, 100])
    result = trans_minor_breaks(t)(major, limits, n=4)
    npt.assert_allclose(
        result,
        [
            1.02961942,
            1.5260563,
            1.85629799,
            2.10413415,
            3.33220451,
            3.8286414,
            4.15888308,
            4.40671925,
        ],
    )


def test_date_breaks():
    # cpython
    x = [datetime(year, 1, 1) for year in [2010, 2026, 2015]]
    limits = min(x), max(x)

    breaks = date_breaks("5 Years")
    years = [d.year for d in breaks(limits)]
    npt.assert_array_equal(years, [2010, 2015, 2020, 2025, 2030])

    breaks = date_breaks("10 Years")
    years = [d.year for d in breaks(limits)]
    npt.assert_array_equal(years, [2010, 2020, 2030])

    # numpy
    x = [np.datetime64(i * 10, "D") for i in range(1, 10)]
    breaks = date_breaks("10 Years")
    limits = min(x), max(x)
    with pytest.raises(AttributeError):
        breaks(limits)

    # NaT
    limits = np.datetime64("NaT"), datetime(2017, 1, 1)
    breaks = date_breaks("10 Years")
    assert len(breaks(limits)) == 0


def test_timedelta_breaks():
    breaks = timedelta_breaks()

    # cpython
    x = [timedelta(days=i * 365) for i in range(25)]
    limits = min(x), max(x)
    major = breaks(limits)
    years = [val.total_seconds() / (365 * 24 * 60 * 60) for val in major]
    npt.assert_array_equal(years, [0, 5, 10, 15, 20, 25])

    x = [timedelta(microseconds=i) for i in range(25)]
    limits = min(x), max(x)
    major = breaks(limits)
    mseconds = [val.total_seconds() * 10**6 for val in major]
    npt.assert_array_equal(mseconds, [0, 5, 10, 15, 20, 25])

    # pandas
    x = [pd.Timedelta(seconds=i * 60) for i in range(10)]
    limits = min(x), max(x)
    major = breaks(limits)
    minutes = [val.total_seconds() / 60 for val in major]
    npt.assert_allclose(minutes, [0, 2, 4, 6, 8])

    # numpy
    x = [np.timedelta64(i * 10, "D") for i in range(1, 10)]
    limits = min(x), max(x)
    with pytest.raises(ValueError):
        breaks(limits)

    # NaT
    limits = pd.NaT, pd.Timedelta(seconds=9 * 60)
    assert len(breaks(limits)) == 0


def test_extended_breaks():
    x = np.arange(100)
    limits = min(x), max(x)
    for n in (5, 7, 10, 13, 31):
        breaks = extended_breaks(n=n)
        assert len(breaks(limits)) <= n + 1

    # Reverse limits
    breaks = extended_breaks(n=7)
    npt.assert_array_equal(breaks((0, 6)), breaks((6, 0)))

    # Infinite limits
    limits = float("-inf"), float("inf")
    breaks = extended_breaks(n=5)
    assert len(breaks(limits)) == 0

    # Zero range discrete
    limits = [1, 1]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]

    # Zero range continuous
    limits = [np.pi, np.pi]
    assert len(breaks(limits)) == 1
    assert breaks(limits)[0] == limits[1]



"""
Scales have guides and these are what help users make sense of
the data mapped onto the scale. Common examples of guides include
the x-axis, the y-axis, the keyed legend and a colorbar legend.
The guides have demarcations(breaks), some of which must be labelled.

The `*_format` functions below create functions that convert data
values as understood by a specific scale and return string
representations of those values. Manipulating the string
representation of a value helps improve readability of the guide.
"""
import re
from bisect import bisect_right

try:
    from zoneinfo import ZoneInfo
except ImportError:
    # python < 3.9
    from backports.zoneinfo import ZoneInfo

import numpy as np

from .breaks import timedelta_helper
from .utils import (
    get_timezone,
    match,
    precision,
    round_any,
    same_log10_order_of_magnitude,
)

__all__ = [
    "comma_format",
    "custom_format",
    "currency_format",
    "dollar_format",
    "percent_format",
    "scientific_format",
    "date_format",
    "mpl_format",
    "log_format",
    "timedelta_format",
    "pvalue_format",
    "ordinal_format",
    "number_bytes_format",
]


class custom_format:
    """
    Custom format

    Parameters
    ----------
    fmt : str, optional
        Format string. Default is the generic new style
        format braces, ``{}``.
    style : 'new' | 'old'
        Whether to use new style or old style formatting.
        New style uses the :meth:`str.format` while old
        style uses ``%``. The format string must be written
        accordingly.

    Examples
    --------
    >>> formatter = custom_format('{:.2f} USD')
    >>> formatter([3.987, 2, 42.42])
    ['3.99 USD', '2.00 USD', '42.42 USD']
    """

    def __init__(self, fmt="{}", style="new"):
        self.fmt = fmt
        self.style = style

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if self.style == "new":
            return [self.fmt.format(val) for val in x]
        elif self.style == "old":
            return [self.fmt % val for val in x]
        else:
            raise ValueError("style should be either 'new' or 'old'")


# formatting functions
class currency_format:
    """
    Currency formatter

    Parameters
    ----------
    prefix : str
        What to put before the value.
    suffix : str
        What to put after the value.
    digits : int
        Number of significant digits
    big_mark : str
        The thousands separator. This is usually
        a comma or a dot.

    Examples
    --------
    >>> x = [1.232, 99.2334, 4.6, 9, 4500]
    >>> currency_format()(x)
    ['$1.23', '$99.23', '$4.60', '$9.00', '$4500.00']
    >>> currency_format('C$', digits=0, big_mark=',')(x)
    ['C$1', 'C$99', 'C$5', 'C$9', 'C$4,500']
    """

    def __init__(self, prefix="$", suffix="", digits=2, big_mark=""):
        self.prefix = prefix
        self.suffix = suffix
        self.digits = digits
        self.big_mark = big_mark

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        # create {:.2f} or {:,.2f}
        big_mark = self.big_mark
        comma = "," if big_mark else ""
        tpl = "".join(
            (
                self.prefix,
                "{:",
                comma,
                ".",
                str(self.digits),
                "f}",
                self.suffix,
            )
        )

        labels = [tpl.format(val) for val in x]
        if big_mark and big_mark != ",":
            labels = [val.replace(",", big_mark) for val in labels]
        return labels


dollar_format = currency_format
dollar = dollar_format()


class comma_format:
    """
    Format number with commas separating thousands

    Parameters
    ----------
    digits : int
        Number of digits after the decimal point.

    Examples
    --------
    >>> comma_format()([1000, 2, 33000, 400])
    ['1,000', '2', '33,000', '400']
    """

    def __init__(self, digits=0):
        self.formatter = currency_format(
            prefix="", digits=digits, big_mark=","
        )

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        return self.formatter(x)


class percent_format:
    """
    Percent formatter

    Multiply by one hundred and display percent sign

    Parameters
    ----------
    use_comma : bool
        If True, use a comma to separate the thousands.
        Default is False.

    Examples
    --------
    >>> formatter = percent_format()
    >>> formatter([.45, 9.515, .01])
    ['45%', '952%', '1%']
    >>> formatter([.654, .8963, .1])
    ['65.4%', '89.6%', '10.0%']
    """

    def __init__(self, use_comma=False):
        self.big_mark = "," if use_comma else ""

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if len(x) == 0:
            return []

        _precision = precision(x)
        x = round_any(x, _precision / 100) * 100

        # When the precision is less than 1, we show
        if _precision > 1:
            digits = 0
        else:
            digits = abs(int(np.log10(_precision)))

        formatter = currency_format(
            prefix="", suffix="%", digits=digits, big_mark=self.big_mark
        )
        labels = formatter(x)
        # Remove unnecessary zeros after the decimal
        pattern = re.compile(r"\.0+%$")
        if all(pattern.search(val) for val in labels):
            labels = [pattern.sub("%", val) for val in labels]
        return labels


percent = percent_format()


class scientific_format:
    """
    Scientific formatter

    Parameters
    ----------
    digits : int
        Significant digits.

    Examples
    --------
    >>> x = [.12, .23, .34, 45]
    >>> scientific_format()(x)
    ['1.2e-01', '2.3e-01', '3.4e-01', '4.5e+01']

    Notes
    -----
    Be careful when using many digits (15+ on a 64
    bit computer). Consider of the `machine epsilon`_.

    .. _machine epsilon: https://en.wikipedia.org/wiki/Machine_epsilon
    """

    def __init__(self, digits=3):
        tpl = "".join(["{:.", str(digits), "e}"])
        self.formatter = custom_format(tpl)

    def __call__(self, x):
        if len(x) == 0:
            return []

        zeros_re = re.compile(r"(0+)e")

        def count_zeros(s):
            match = zeros_re.search(s)
            if match:
                return len(match.group(1))
            else:
                return 0

        # format and then remove superfluous zeros
        labels = self.formatter(x)
        n = min([count_zeros(val) for val in labels])
        if n:
            labels = [val.replace("0" * n + "e", "e") for val in labels]
        return labels


scientific = scientific_format()


def _format(formatter, x):
    """
    Helper to format and tidy up
    """
    # For MPL to play nice
    formatter.create_dummy_axis()
    # For sensible decimal places
    formatter.set_locs([val for val in x if ~np.isnan(val)])
    try:
        oom = int(formatter.orderOfMagnitude)
    except AttributeError:
        oom = 0
    labels = [formatter(tick) for tick in x]

    # Remove unnecessary decimals
    pattern = re.compile(r"\.0+$")
    for i, label in enumerate(labels):
        match = pattern.search(label)
        if match:
            labels[i] = pattern.sub("", label)

    # MPL does not add the exponential component
    if oom:
        labels = ["{}e{}".format(s, oom) if s != "0" else s for s in labels]
    return labels


class mpl_format:
    """
    Format using MPL formatter for scalars

    Examples
    --------
    >>> mpl_format()([.654, .8963, .1])
    ['0.6540', '0.8963', '0.1000']
    """

    def __init__(self):
        from matplotlib.ticker import ScalarFormatter

        self.formatter = ScalarFormatter(useOffset=False)

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        return _format(self.formatter, x)


class log_format:
    """
    Log Formatter

    Parameters
    ----------
    base : int
        Base of the logarithm. Default is 10.
    exponent_limits : tuple
        limits (int, int) where if the any of the powers of the
        numbers falls outside, then the labels will be in
        exponent form. This only applies for base 10.
    mathtex : bool
        If True, return the labels in mathtex format as understood
        by Matplotlib.

    Examples
    --------
    >>> log_format()([0.001, 0.1, 100])
    ['0.001', '0.1', '100']

    >>> log_format()([0.0001, 0.1, 10000])
    ['1e-4', '1e-1', '1e4']

    >>> log_format(mathtex=True)([0.0001, 0.1, 10000])
    ['$10^{-4}$', '$10^{-1}$', '$10^{4}$']
    """

    def __init__(self, base=10, exponent_limits=(-4, 4), mathtex=False):
        self.base = base
        self.exponent_limits = exponent_limits
        self.mathtex = mathtex

    def _tidyup_labels(self, labels):
        """
        Make all labels uniform in format

        Remove redundant zeros for labels in exponential format.

        Parameters
        ----------
        labels : list-like
            Labels to be tidied.

        Returns
        -------
        out : list-like
            Labels
        """

        def remove_zeroes(s):
            """
            Remove unnecessary zeros for float string s
            """
            tup = s.split("e")
            if len(tup) == 2:
                mantissa = tup[0].rstrip("0").rstrip(".")
                exponent = int(tup[1])
                if exponent:
                    s = "%se%d" % (mantissa, exponent)
                else:
                    s = mantissa
            return s

        def as_exp(s):
            """
            Float string s as in exponential format
            """
            return s if "e" in s else "{:1.0e}".format(float(s))

        def as_mathtex(s):
            """
            Mathtex for maplotlib
            """
            if "e" not in s:
                assert s == "1", f"Unexpected value {s = }, instead of '1'"
                return f"${self.base}^{{0}}$"

            exp = s.split("e")[1]
            return f"${self.base}^{{{exp}}}$"

        # If any are in exponential format, make all of
        # them expontential
        has_e = ["e" in x for x in labels]
        if not all(has_e) and sum(has_e):
            labels = [as_exp(x) for x in labels]

        labels = [remove_zeroes(x) for x in labels]

        has_e = ["e" in x for x in labels]
        if self.mathtex and any(has_e):
            labels = [as_mathtex(x) for x in labels]

        return labels

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        if len(x) == 0:
            return []

        # Decide on using exponents
        if self.base == 10:
            xmin = int(np.floor(np.log10(np.min(x))))
            xmax = int(np.ceil(np.log10(np.max(x))))
            emin, emax = self.exponent_limits
            all_multiples = np.all([np.log10(num).is_integer() for num in x])
            # Order of magnitude of the minimum and maximum
            if same_log10_order_of_magnitude(x):
                f = mpl_format()
                f.formatter.set_powerlimits((emin, emax))
                return f(x)
            elif all_multiples and (xmin <= emin or xmax >= emax):
                fmt = "{:1.0e}"
            else:
                fmt = "{:g}"
            labels = [fmt.format(num) for num in x]
            return self._tidyup_labels(labels)
        else:

            def _exp(num, base):
                e = np.log(num) / np.log(base)
                e_round = np.round(e)
                if np.isclose(e, e_round):
                    e = int(e_round)
                else:
                    e = np.round(e, 3)
                return e

            base_txt = f"{self.base}"
            if self.base == np.e:
                base_txt = "e"

            if self.mathtex:
                fmt_parts = (f"${base_txt}^", "{{{e}}}$")
            else:
                fmt_parts = (f"{base_txt}^", "{e}")

            fmt = "".join(fmt_parts)
            exps = [_exp(num, self.base) for num in x]
            labels = [fmt.format(e=e) for e in exps]
            return labels


class date_format:
    """
    Datetime formatter

    Parameters
    ----------
    fmt : str
        Format string. See
        :ref:`strftime <strftime-strptime-behavior>`.
    tz : datetime.tzinfo, optional
        Time zone information. If none is specified, the
        time zone will be that of the first date. If the
        first date has no time information then a time zone
        is chosen by other means.

    Examples
    --------
    >>> from datetime import datetime
    >>> x = [datetime(x, 1, 1) for x in [2010, 2014, 2018, 2022]]
    >>> date_format()(x)
    ['2010-01-01', '2014-01-01', '2018-01-01', '2022-01-01']
    >>> date_format('%Y')(x)
    ['2010', '2014', '2018', '2022']

    Can format time

    >>> x = [datetime(2017, 12, 1, 16, 5, 7)]
    >>> date_format("%Y-%m-%d %H:%M:%S")(x)
    ['2017-12-01 16:05:07']

    Time zones are respected

    >>> UTC = ZoneInfo('UTC')
    >>> UG = ZoneInfo('Africa/Kampala')
    >>> x = [datetime(2010, 1, 1, i) for i in [8, 15]]
    >>> x_tz = [datetime(2010, 1, 1, i, tzinfo=UG) for i in [8, 15]]
    >>> date_format('%Y-%m-%d %H:%M')(x)
    ['2010-01-01 08:00', '2010-01-01 15:00']
    >>> date_format('%Y-%m-%d %H:%M')(x_tz)
    ['2010-01-01 08:00', '2010-01-01 15:00']

    Format with a specific time zone

    >>> date_format('%Y-%m-%d %H:%M', tz=UTC)(x_tz)
    ['2010-01-01 05:00', '2010-01-01 12:00']
    >>> date_format('%Y-%m-%d %H:%M', tz='EST')(x_tz)
    ['2010-01-01 00:00', '2010-01-01 07:00']
    """

    def __init__(self, fmt="%Y-%m-%d", tz=None):
        if isinstance(tz, str):
            tz = ZoneInfo(tz)
        from matplotlib.dates import DateFormatter

        self.formatter = DateFormatter(fmt, tz=tz)
        self.tz = tz

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        from matplotlib.dates import date2num

        # Formatter timezone
        if self.tz is None and len(x):
            self.formatter.tz = get_timezone(x)

        # The formatter is tied to axes and takes
        # breaks in ordinal format.
        x = [date2num(val) for val in x]
        return _format(self.formatter, x)


class timedelta_format:
    """
    Timedelta formatter

    Parameters
    ----------
    units : str, optional
        The units in which the breaks will be computed.
        If None, they are decided automatically. Otherwise,
        the value should be one of::

            'ns'    # nanoseconds
            'us'    # microseconds
            'ms'    # milliseconds
            's'     # secondss
            'm'     # minute
            'h'     # hour
            'd'     # day
            'w'     # week
            'M'     # month
            'y'     # year

    add_units : bool
        Whether to append the units identifier string
        to the values.
    usetext : bool
        If True, they microseconds identifier string is
        rendered with greek letter *mu*. Default is False.

    Examples
    --------
    >>> from datetime import timedelta
    >>> x = [timedelta(days=31*i) for i in range(5)]
    >>> timedelta_format()(x)
    ['0', '1 month', '2 months', '3 months', '4 months']
    >>> timedelta_format(units='d')(x)
    ['0', '31 days', '62 days', '93 days', '124 days']
    >>> timedelta_format(units='d', add_units=False)(x)
    ['0', '31', '62', '93', '124']
    """

    abbreviations = {
        "ns": "ns",
        "us": "us",
        "ms": "ms",
        "s": "s",
        "m": " minute",
        "h": " hour",
        "d": " day",
        "w": " week",
        "M": " month",
        "y": " year",
    }

    def __init__(self, units=None, add_units=True, usetex=False):
        self.units = units
        self.add_units = add_units
        self.usetex = usetex
        self._mpl_format = mpl_format()

    def __call__(self, x):
        if len(x) == 0:
            return []

        labels = []
        values, _units = timedelta_helper.format_info(x, self.units)
        plural = "" if _units.endswith("s") else "s"
        ulabel = self.abbreviations[_units]
        if ulabel == "us" and self.usetex:
            ulabel = r"$\mu s$"
        _labels = self._mpl_format(values)

        if not self.add_units:
            return _labels

        for num, num_label in zip(values, _labels):
            s = "" if num == 1 else plural
            # 0 has no units
            _ulabel = "" if num == 0 else ulabel + s
            labels.append("".join([num_label, _ulabel]))

        return labels


class pvalue_format:
    """
    p-values Formatter

    Parameters
    ----------
    accuracy : float
        Number to round to
    add_p : bool
        Whether to prepend "p=" or "p<" to the output

    Examples
    --------
    >>> x = [.90, .15, .015, .009, 0.0005]
    >>> pvalue_format()(x)
    ['0.9', '0.15', '0.015', '0.009', '<0.001']
    >>> pvalue_format(0.1)(x)
    ['0.9', '0.1', '<0.1', '<0.1', '<0.1']
    >>> pvalue_format(0.1, True)(x)
    ['p=0.9', 'p=0.1', 'p<0.1', 'p<0.1', 'p<0.1']
    """

    def __init__(self, accuracy=0.001, add_p=False):
        self.accuracy = accuracy
        self.add_p = add_p

    def __call__(self, x):
        """
        Format a sequence of inputs

        Parameters
        ----------
        x : array
            Input

        Returns
        -------
        out : list
            List of strings.
        """
        x = round_any(x, self.accuracy)
        below = [num < self.accuracy for num in x]

        if self.add_p:
            eq_fmt = "p={:g}".format
            below_label = "p<{:g}".format(self.accuracy)
        else:
            eq_fmt = "{:g}".format
            below_label = "<{:g}".format(self.accuracy)

        labels = [below_label if b else eq_fmt(i) for i, b in zip(x, below)]
        return labels


def ordinal(n, prefix="", suffix="", big_mark=""):
    # General Case: 0th, 1st, 2nd, 3rd, 4th, 5th, 6th, 7th, 8th, 9th
    # Special Case: 10th, 11th, 12th, 13th
    n = int(n)
    idx = np.min((n % 10, 4))
    _suffix = ["th", "st", "nd", "rd", "th"][idx]
    if 11 <= (n % 100) <= 13:
        _suffix = "th"

    if big_mark:
        s = "{:,}".format(n)
        if big_mark != ",":
            s = s.replace(",", big_mark)
    else:
        s = "{}".format(n)

    return "{}{}{}{}".format(prefix, s, _suffix, suffix)


class ordinal_format:
    """
    Ordinal Formatter

    Parameters
    ----------
    prefix : str
        What to put before the value.
    suffix : str
        What to put after the value.
    big_mark : str
        The thousands separator. This is usually
        a comma or a dot.

    Examples
    --------
    >>> ordinal_format()(range(8))
    ['0th', '1st', '2nd', '3rd', '4th', '5th', '6th', '7th']
    >>> ordinal_format(suffix=' Number')(range(11, 15))
    ['11th Number', '12th Number', '13th Number', '14th Number']
    """

    def __init__(self, prefix="", suffix="", big_mark=""):
        self.prefix = prefix
        self.suffix = suffix
        self.big_mark = big_mark

    def __call__(self, x):
        labels = [
            ordinal(num, self.prefix, self.suffix, self.big_mark) for num in x
        ]
        return labels


class number_bytes_format:
    """
    Bytes Formatter

    Parameters
    ----------
    symbol : str
        Valid symbols are "B", "kB", "MB", "GB", "TB", "PB", "EB",
        "ZB", and "YB" for SI units, and the "iB" variants for
        binary units. Default is "auto" where the symbol to be
        used is determined separately for each value of 1x.
    units : "binary" | "si"
        Which unit base to use, 1024 for "binary" or 1000 for "si".
    fmt : str, optional
        Format sting. Default is ``{:.0f}``.

    Examples
    --------
    >>> x = [1000, 1000000, 4e5]
    >>> number_bytes_format()(x)
    ['1000 B', '977 KiB', '391 KiB']
    >>> number_bytes_format(units='si')(x)
    ['1 kB', '1 MB', '400 kB']
    """

    def __init__(self, symbol="auto", units="binary", fmt="{:.0f} "):
        self.symbol = symbol
        self.units = units
        self.fmt = fmt

        if units == "si":
            self.base = 1000
            self._all_symbols = [
                "B",
                "kB",
                "MB",
                "GB",
                "TB",
                "PB",
                "EB",
                "ZB",
                "YB",
            ]
        else:
            self.base = 1024
            self._all_symbols = [
                "B",
                "KiB",
                "MiB",
                "GiB",
                "TiB",
                "PiB",
                "EiB",
                "ZiB",
                "YiB",
            ]

        # possible exponents of base: eg 1000^1, 1000^2, 1000^3, ...
        exponents = np.arange(1, len(self._all_symbols) + 1, dtype=float)
        self._powers = self.base**exponents
        self._validate_symbol(symbol, ["auto"] + self._all_symbols)

    def __call__(self, x):
        _all_symbols = self._all_symbols
        symbol = self.symbol
        if symbol == "auto":
            power = [bisect_right(self._powers, val) for val in x]
            symbols = [_all_symbols[p] for p in power]
        else:
            power = np.array(match([symbol], _all_symbols))
            symbols = [symbol] * len(x)

        x = np.asarray(x)
        power = np.asarray(power, dtype=float)
        values = x / self.base**power
        fmt = (self.fmt + "{}").format
        labels = [fmt(v, s) for v, s in zip(values, symbols)]
        return labels

    def _validate_symbol(self, symbol, allowed_symbols):
        if symbol not in allowed_symbols:
            raise ValueError(
                "Symbol must be one of {}".format(allowed_symbols)
            )



"""
According to *On the theory of scales of measurement* by **S.S. Stevens**,
scales can be classified in four ways -- *nominal*, *ordinal*,
*interval* and *ratio*. Using current(2016) terminology, *nominal* data
is made up of unordered categories, *ordinal* data is made up of ordered
categories and the two can be classified as *discrete*. On the other hand
both *interval* and *ratio* data are *continuous*.

The scale classes below show how the rest of the Mizani package can be
used to implement the two categories of scales. The key tasks are
*training* and *mapping* and these correspond to the **train** and
**map** methods.

To train a scale on data means, to make the scale learn the limits of
the data. This is elaborate (or worthy of a dedicated method) for two
reasons:

    - *Practical* -- data may be split up across more than one object,
      yet all will be represented by a single scale.
    - *Conceptual* -- training is a key action that may need to be inserted
      into multiple locations of the data processing pipeline before a
      graphic can be created.

To map data onto a scale means, to associate data values with
values(potential readings) on a scale. This is perhaps the most important
concept unpinning a scale.

The **apply** methods are simple examples of how to put it all together.
"""
import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from .bounds import censor, rescale
from .utils import (
    CONTINUOUS_KINDS,
    DISCRETE_KINDS,
    get_categories,
    match,
    min_max,
)

__all__ = ["scale_continuous", "scale_discrete"]


class scale_continuous:
    """
    Continuous scale
    """

    @classmethod
    def apply(cls, x, palette, na_value=None, trans=None):
        """
        Scale data continuously

        Parameters
        ----------
        x : array_like
            Continuous values to scale
        palette : callable ``f(x)``
            Palette to use
        na_value : object
            Value to use for missing values.
        trans : trans
            How to transform the data before scaling. If
            ``None``, no transformation is done.

        Returns
        -------
        out : array_like
            Scaled values
        """
        if trans is not None:
            x = trans.transform(x)

        limits = cls.train(x)
        return cls.map(x, palette, limits, na_value)

    @classmethod
    def train(cls, new_data, old=None):
        """
        Train a continuous scale

        Parameters
        ----------
        new_data : array_like
            New values
        old : array_like
            Old range. Most likely a tuple of length 2.

        Returns
        -------
        out : tuple
            Limits(range) of the scale
        """
        if not len(new_data):
            return old

        if not hasattr(new_data, "dtype"):
            new_data = np.asarray(new_data)

        if new_data.dtype.kind not in CONTINUOUS_KINDS:
            raise TypeError("Discrete value supplied to continuous scale")

        if old is not None:
            new_data = np.hstack([new_data, old])

        return min_max(new_data, na_rm=True, finite=True)

    @classmethod
    def map(cls, x, palette, limits, na_value=None, oob=censor):
        """
        Map values to a continuous palette

        Parameters
        ----------
        x : array_like
            Continuous values to scale
        palette : callable ``f(x)``
            palette to use
        na_value : object
            Value to use for missing values.
        oob : callable ``f(x)``
            Function to deal with values that are
            beyond the limits

        Returns
        -------
        out : array_like
            Values mapped onto a palette
        """
        x = oob(rescale(x, _from=limits))
        pal = palette(x)
        try:
            pal[pd.isnull(x)] = na_value
        except TypeError:
            pal = [v if not pd.isnull(v) else na_value for v in pal]

        return pal


class scale_discrete:
    """
    Discrete scale
    """

    @classmethod
    def apply(cls, x, palette, na_value=None):
        """
        Scale data discretely

        Parameters
        ----------
        x : array_like
            Discrete values to scale
        palette : callable ``f(x)``
            Palette to use
        na_value : object
            Value to use for missing values.

        Returns
        -------
        out : array_like
            Scaled values
        """
        limits = cls.train(x)
        return cls.map(x, palette, limits, na_value)

    @classmethod
    def train(cls, new_data, old=None, drop=False, na_rm=False):
        """
        Train a continuous scale

        Parameters
        ----------
        new_data : array_like
            New values
        old : array_like
            Old range. List of values known to the scale.
        drop : bool
            Whether to drop(not include) unused categories
        na_rm : bool
            If ``True``, remove missing values. Missing values
            are either ``NaN`` or ``None``.

        Returns
        -------
        out : list
            Values covered by the scale
        """
        if not len(new_data):
            return old

        if old is None:
            old = []
        else:
            old = list(old)

        # Get the missing values (NaN & Nones) locations and remove them
        nan_bool_idx = pd.isnull(new_data)
        has_na = np.any(nan_bool_idx)
        if not hasattr(new_data, "dtype"):
            new_data = np.asarray(new_data)
        new_data = new_data[~nan_bool_idx]

        if new_data.dtype.kind not in DISCRETE_KINDS:
            raise TypeError("Continuous value supplied to discrete scale")

        # Train i.e. get the new values
        if pdtypes.is_categorical_dtype(new_data):
            categories = get_categories(new_data)
            if drop:
                present = set(new_data.drop_duplicates())
                new = [i for i in categories if i in present]
            else:
                new = list(categories)
        else:
            new = np.unique(new_data)
            new.sort()

        # update old
        old_set = set(old)
        if pdtypes.is_categorical_dtype(new_data):
            # The limits are in the order of the categories
            all_set = old_set | set(new)
            ordered_cats = categories.union(old, sort=False)
            limits = [c for c in ordered_cats if c in all_set]
        else:
            limits = old + [i for i in new if (i not in old_set)]

        # Add nan if required
        has_na_limits = any(pd.isnull(limits))
        if not has_na_limits and not na_rm and has_na:
            limits.append(np.nan)
        return limits

    @classmethod
    def map(cls, x, palette, limits, na_value=None):
        """
        Map values to a discrete palette

        Parameters
        ----------
        palette : callable ``f(x)``
            palette to use
        x : array_like
            Continuous values to scale
        na_value : object
            Value to use for missing values.

        Returns
        -------
        out : array_like
            Values mapped onto a palette
        """
        n = len(limits)
        pal = palette(n)[match(x, limits)]
        try:
            pal[pd.isnull(x)] = na_value
        except TypeError:
            pal = [v if not pd.isnull(v) else na_value for v in pal]

        return pal



print("import: 'mizani'")
import mizani

print("import: 'mizani.external'")
import mizani.external

print("import: 'mizani.tests'")
import mizani.tests

print("import: 'mizani.colors'")
import mizani.colors




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from types import ModuleType

    from mizani.typing import ColorPalette, ColorScheme, ColorSchemeShort

__all__ = ["get_palette_names"]


def get_palette_names(scheme: ColorScheme | ColorSchemeShort) -> list[str]:
    """
    Return list of palette names
    """
    mod = get_palette_module(scheme)
    names = mod.__all__
    return names.copy()


def get_palette_module(scheme: ColorScheme | ColorSchemeShort) -> ModuleType:
    """
    Return Module with the palettes for the scheme
    """
    if scheme in ("sequential", "seq"):
        from . import sequential

        return sequential
    elif scheme in ("qualitative", "qual"):
        from . import qualitative

        return qualitative
    elif scheme in ("diverging", "div"):
        from . import diverging

        return diverging
    else:
        raise ValueError(f"Unknown type of brewer palette: {type}")


def number_to_name(scheme: ColorScheme | ColorSchemeShort, n: int) -> str:
    """
    Return palette name that corresponds to a given number

    Uses alphabetical ordering
    """
    mod = get_palette_module(scheme)
    names = mod.__all__
    if n > len(names):
        raise ValueError(
            f"There are only '{n}' palettes of type {scheme}. "
            f"You requested palette no. {n}"
        )
    return names[n - 1]


def get_color_palette(
    scheme: ColorScheme | ColorSchemeShort, palette: int | str
) -> ColorPalette:
    """
    Return color palette from a given scheme
    """
    if isinstance(palette, int):
        palette = number_to_name(scheme, palette)
    mod = get_palette_module(scheme)
    return getattr(mod, palette)



import numpy as np
import pandas as pd
import pytest

from mizani.utils import (
    first_element,
    get_categories,
    match,
    min_max,
    multitype_sort,
    precision,
    round_any,
    same_log10_order_of_magnitude,
)


def test_round_any():
    x = 4.632
    assert round_any(x, 1) == 5
    assert round_any(x, 2) == 4
    assert round_any(x, 3) == 6
    assert round_any(x, 4) == 4
    assert round_any(x, 5) == 5
    assert round_any(x, 1.5) == 4.5

    # Maintains the same index
    s = pd.Series([1.1, 2.2, 3.3], index=[3, 2, 1])
    result = round_any(s, 2)
    assert s.index.equals(result.index)


def test_min_max():
    x = [1, 2, 3, 4, 5]
    _min, _max = min_max(x)
    assert _min == 1
    assert _max == 5

    x = [1, float("-inf"), 3, 4, 5]
    _min, _max = min_max(x)
    assert _min == 1
    assert _max == 5

    _min, _max = min_max(x, finite=False)
    assert _min == float("-inf")
    assert _max == 5

    x = [1, 2, float("nan"), 4, 5]
    _min, _max = min_max(x, na_rm=True)
    assert _min == 1
    assert _max == 5

    x = [1, 2, float("nan"), 4, 5, float("inf")]
    _min, _max = min_max(x, na_rm=True, finite=False)
    assert _min == 1
    assert _max == float("inf")

    _min, _max = min_max(x)
    assert str(_min) == "nan"
    assert str(_max) == "nan"

    x = [float("nan"), float("nan"), float("nan")]
    _min, _max = min_max(x, na_rm=True)
    assert _min == float("-inf")
    assert _max == float("inf")


def test_match():
    v1 = [0, 1, 2, 3, 4, 5]
    v2 = [5, 4, 3, 2, 1, 0]
    result = match(v1, v2)
    assert result == v2

    # Positions of the first match
    result = match(v1, v2 + v2)
    assert result == v2

    result = match(v1, v2, incomparables=[1, 2])
    assert result == [5, -1, -1, 2, 1, 0]

    result = match(v1, v2, start=1)
    assert result == [6, 5, 4, 3, 2, 1]

    v2 = [5, 99, 3, 2, 1, 0]
    result = match(v1, v2)
    assert result == [5, 4, 3, 2, -1, 0]


def test_precision():
    assert precision(0.0037) == 0.001
    assert precision(0.5) == 0.1
    assert precision(9) == 1
    assert precision(24) == 10
    assert precision(784) == 100
    assert precision([0, 0]) == 1


def test_first_element():
    x = [3, 4, 5]
    s = pd.Series(x)
    a = np.array([3, 4, 5])

    assert first_element(x) == 3
    assert first_element(s) == 3
    assert first_element(s[1:]) == 4
    assert first_element(a) == 3
    assert first_element(a[1:]) == 4

    with pytest.raises(StopIteration):
        first_element([])

    with pytest.raises(RuntimeError):
        first_element(iter(x))


def test_multitype_sort():
    a = ["c", float("nan"), 1, "b", "a", 2.0, 0]
    result = multitype_sort(a)
    # Any consecutive elements of the sametype are
    # sorted
    for i, x in enumerate(result[1:], start=1):
        x_prev = result[i - 1]
        if type(x_prev) is type(x):
            # cannot compare nan with anything
            if isinstance(x, float) and (np.isnan(x_prev) or np.isnan(x)):
                continue
            assert x_prev <= x


def test_same_log10_order_of_magnitude():
    # Default delta
    assert same_log10_order_of_magnitude((2, 8))
    assert same_log10_order_of_magnitude((35, 80.8))
    assert same_log10_order_of_magnitude((232.3, 730))

    assert not same_log10_order_of_magnitude((1, 18))
    assert not same_log10_order_of_magnitude((35, 800))
    assert not same_log10_order_of_magnitude((32, 730))

    assert not same_log10_order_of_magnitude((1, 9.9))
    assert not same_log10_order_of_magnitude((35, 91))
    assert not same_log10_order_of_magnitude((232.3, 950))

    # delta = 0
    assert same_log10_order_of_magnitude((1, 9.9), delta=0)
    assert same_log10_order_of_magnitude((35, 91), delta=0)
    assert same_log10_order_of_magnitude((232.3, 950), delta=0)


def test_get_categories():
    lst = list("abcd")
    s = pd.Series(lst)
    c = pd.Categorical(lst)
    sc = pd.Series(c)

    categories = pd.Index(lst)
    assert categories.equals(get_categories(c))
    assert categories.equals(get_categories(sc))

    with pytest.raises(TypeError):
        assert categories.equals(get_categories(lst))

    with pytest.raises(TypeError):
        assert categories.equals(get_categories(s))



import warnings
from datetime import datetime, timedelta

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt
import pytest
from pytest import approx

from mizani.bounds import (
    censor,
    expand_range,
    expand_range_distinct,
    rescale,
    rescale_max,
    rescale_mid,
    squish,
    squish_infinite,
    zero_range,
)

NaT_type = type(pd.NaT)


def test_censor():
    x = list(range(10))
    xx = censor(x, (2, 8))
    assert np.isnan(xx[0])
    assert np.isnan(xx[1])
    assert np.isnan(xx[9])

    df = pd.DataFrame({"x": x, "y": range(10)})
    df["x"] = censor(df["x"], (2, 8))
    assert np.isnan(df["x"][0])
    assert np.isnan(df["x"][1])
    assert np.isnan(df["x"][9])

    df["y"] = censor(df["y"], (-2, 18))
    assert issubclass(df["y"].dtype.type, np.integer)

    # datetime
    limits = datetime(2010, 1, 1), datetime(2020, 1, 1)
    x = [datetime(year, 1, 1) for year in range(2008, 2023)]
    result = censor(x, limits)
    assert result[2:-2] == x[2:-2]
    assert result[:2] == ["NaT", "NaT"]
    assert result[-2:] == ["NaT", "NaT"]

    # timedelta
    limits = timedelta(seconds=2010), timedelta(seconds=2020)
    x = [timedelta(seconds=i) for i in range(2008, 2023)]
    result = censor(x, limits)
    assert result[2:-2] == x[2:-2]
    assert result[:2] == ["NaT", "NaT"]
    assert result[-2:] == ["NaT", "NaT"]

    # pd.timestamp
    limits = pd.Timestamp(200 * 1e16), pd.Timestamp(205 * 1e16)
    x = [pd.Timestamp(i * 1e16) for i in range(198, 208)]
    result = censor(x, limits)
    assert result[2:-2] == x[2:-2]
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    x1 = np.array(x)
    result = censor(x1, limits)
    npt.assert_array_equal(result[2:-2], x1[2:-2])
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    x2 = pd.Series(x)
    result = censor(x2, limits)
    pdt.assert_series_equal(result[2:-2], x2[2:-2])
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    # np.datetime
    limits = np.datetime64(200, "D"), np.datetime64(205, "D")
    x = [np.datetime64(i, "D") for i in range(198, 208)]
    x2 = np.array(x)
    result = censor(x2, limits)
    npt.assert_array_equal(result[2:-2], x2[2:-2])
    assert all(isinstance(val, np.datetime64) for val in result[:2])
    assert all(isinstance(val, np.datetime64) for val in result[-2:])

    # pd.Timedelta
    limits = pd.Timedelta(seconds=2010), pd.Timedelta(seconds=2020)
    x = [pd.Timedelta(seconds=i) for i in range(2008, 2023)]
    result = censor(x, limits)
    assert isinstance(result, list)
    assert result[2:-2] == x[2:-2]
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    x4 = np.array(x)
    result = censor(x4, limits)
    npt.assert_array_equal(result[2:-2], x4[2:-2])
    assert all(isinstance(val, NaT_type) for val in result[:2])
    assert all(isinstance(val, NaT_type) for val in result[-2:])

    # np.timedelta64
    limits = np.timedelta64(200, "D"), np.timedelta64(205, "D")
    x = [np.timedelta64(i, "D") for i in range(198, 208)]
    x5 = np.array(x)
    result = censor(x5, limits)
    npt.assert_array_equal(result[2:-2], x5[2:-2])
    assert all(isinstance(val, np.timedelta64) for val in result[:2])
    assert all(isinstance(val, np.timedelta64) for val in result[-2:])

    # branches #
    x = np.array([1, 2, np.inf, 3, 4, 11])
    result = censor(x, (0, 10), only_finite=False)
    npt.assert_array_equal(result, np.array([1, 2, np.nan, 3, 4, np.nan]))

    result = censor([], (-2, 18))
    assert len(result) == 0

    with pytest.raises(ValueError):
        result = censor(["a", "b", "c"], ("a", "z"))


def test_expand_range():
    assert expand_range((0, 1)) == (0, 1)
    assert expand_range((0, 1), mul=2) == (-2, 3)
    assert expand_range((0, 1), add=2) == (-2, 3)
    assert expand_range((0, 1), mul=2, add=2) == (-4, 5)
    assert expand_range((1, 1), mul=2, add=2, zero_width=1) == (0.5, 1.5)
    assert expand_range(0) == (-0.5, 0.5)

    def diff(x):
        return x[1] - x[0]

    # datetime
    one_day = datetime(2010, 1, 2) - datetime(2010, 1, 1)
    limits = datetime(2010, 1, 1), datetime(2010, 1, 2)
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    limits = datetime(2010, 1, 1), datetime(2010, 1, 1)  # zero range
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # pd.Timestamp
    one_day = pd.Timestamp("2010-01-02") - pd.Timestamp("2010-01-01")
    limits = pd.Timestamp("2010-01-01"), pd.Timestamp("2010-01-02")
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = pd.Timestamp("2010-01-01"), pd.Timestamp("2010-01-01")
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # np.datetime64
    one_day = np.datetime64(1, "D") - np.datetime64(0, "D")
    limits = np.datetime64(14610, "D"), np.datetime64(14611, "D")
    result = expand_range(limits, add=one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = np.datetime64(14610, "D"), np.datetime64(14611, "D")
    result = expand_range(limits, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # timedelta
    one_day = timedelta(days=1)
    limits = timedelta(days=1), timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = timedelta(days=10), timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # pd.Timedelta
    one_day = pd.Timedelta(days=1)
    limits = pd.Timedelta(days=1), pd.Timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = pd.Timedelta(days=10), pd.Timedelta(days=10)
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day

    # timedelta64
    one_day = np.timedelta64(1, "D")
    limits = np.timedelta64(1, "D"), np.timedelta64(10, "D")
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 2 * one_day

    result = expand_range(limits, mul=0.5, add=one_day)
    diff(result) == 2 * diff(limits) + 2 * one_day

    limits = np.timedelta64(1, "D"), np.timedelta64(1, "D")
    result = expand_range(limits, add=one_day, zero_width=30 * one_day)
    diff(result) == diff(limits) + 30 * one_day


def test_expand_range_distinct():
    assert expand_range_distinct((0, 1)) == (0, 1)
    assert expand_range_distinct((0, 1), (2, 0)) == (-2, 3)
    assert expand_range_distinct((0, 1), (2, 0, 2, 0)) == (-2, 3)
    assert expand_range_distinct((0, 1), (0, 2)) == (-2, 3)
    assert expand_range_distinct((0, 1), (0, 2, 0, 2)) == (-2, 3)
    assert expand_range_distinct((0, 1), (2, 2, 2, 2)) == (-4, 5)
    assert expand_range_distinct((1, 1), (2, 2), zero_width=1) == (0.5, 1.5)


def test_rescale():
    # [0, 10] -> [0, 1]
    # Results are invariant to uniformly translated
    # or expanded inputs
    a = np.arange(0, 11)
    npt.assert_allclose(rescale(a), a * 0.1)
    npt.assert_allclose(rescale(a), rescale(a - 42))
    npt.assert_allclose(rescale(a), rescale(a + 42))
    npt.assert_allclose(rescale(a), rescale(a / np.pi))
    npt.assert_allclose(rescale(a), rescale(a * np.pi))

    # Some more
    n = 6
    a = np.arange(0, n)
    npt.assert_allclose(rescale(a), a / (n - 1))
    npt.assert_allclose(rescale(a, _from=(0, 10)), a * 0.1)
    npt.assert_allclose(rescale(a, to=(0, n * (n - 1))), a * n)


def test_rescale_max():
    a = np.arange(0, 11)
    # Max & first element has no effect
    assert max(rescale_max(a, to=(0, 5))) == 5
    assert max(rescale_max(a, to=(4, 5))) == 5

    # Expanded & first elements have no effect
    assert max(rescale_max(a, to=(0, 5), _from=(0, 3))) > 5
    assert max(rescale_max(a, to=(2, 5), _from=(2, 3))) > 5

    # branches #
    assert rescale_max(2, _from=(0, 10)) == 0.2

    # Maintains the same index
    s = pd.Series([1, 2, 3], index=[3, 2, 1])
    result = rescale_max(s)
    assert s.index.equals(result.index)

    x = [-5, -4, -3, -2, -1, 0]
    result = rescale_max(x)
    assert result == approx([0, 0.2, 0.4, 0.6, 0.8, 1.0])

    result = rescale_max(x, to=(0, 5))
    assert result == approx([0, 1, 2, 3, 4, 5])


def test_rescale_mid():
    a = [1, 2, 3]
    # no change
    npt.assert_allclose(rescale_mid(a, to=(1, 3), mid=2), a)
    npt.assert_allclose(rescale_mid(a, to=(1, 4), _from=(1, 4), mid=2.5), a)

    npt.assert_allclose(rescale_mid(a, mid=1), [0.5, 0.75, 1])
    npt.assert_allclose(rescale_mid(a, mid=2), [0, 0.5, 1])
    npt.assert_allclose(rescale_mid(a, mid=3), [0, 0.25, 0.5])

    # branches #
    npt.assert_allclose(rescale_mid(2, to=(1, 3), mid=2), 2)
    npt.assert_allclose(rescale_mid([2], _from=(2, 2), to=(2, 2), mid=2), [2])

    # Maintains the same index
    s = pd.Series([1, 2, 3], index=[3, 2, 1])
    result = rescale_mid(s, mid=1)
    assert s.index.equals(result.index)


def test_squish_infinite():
    a = [-np.inf, np.inf, -np.inf, np.inf]
    npt.assert_allclose(squish_infinite(a), [0, 1, 0, 1])
    npt.assert_allclose(
        squish_infinite(a, (-100, 100)), [-100, 100, -100, 100]
    )

    b = np.array([5, -np.inf, 2, 3, 6])
    npt.assert_allclose(squish_infinite(b, (1, 10)), [5, 1, 2, 3, 6])

    # Maintains the same index
    s = pd.Series([1, 2, 3, 4], index=[4, 3, 2, 1])
    result = squish_infinite(s)
    assert s.index.equals(result.index)


def test_squish():
    a = [-np.inf, np.inf, -np.inf, np.inf]
    npt.assert_allclose(squish(a, only_finite=False), [0, 1, 0, 1])
    npt.assert_allclose(squish(a, only_finite=True), a)
    npt.assert_allclose(
        squish(a, (-100, 100), only_finite=False), [-100, 100, -100, 100]
    )

    b = np.array([5, 0, -2, 3, 10])
    npt.assert_allclose(squish(b, (0, 5)), [5, 0, 0, 3, 5])

    c = np.array([5, -np.inf, 2, 3, 6])
    npt.assert_allclose(squish(c, (1, 10), only_finite=False), [5, 1, 2, 3, 6])
    npt.assert_allclose(squish(c, (1, 10)), c)

    # Maintains the same index
    s = pd.Series([0.1, 0.2, 0.3, 9], index=[4, 3, 2, 1])
    result = squish(s)
    assert s.index.equals(result.index)


def test_zero_range():
    c = np.array
    eps = np.finfo(float).eps

    assert zero_range(c((1, 1 + eps)))
    assert zero_range(c((1, 1 + 99 * eps)))

    # Crossed the tol threshold
    # assert zero_range(c((1, 1 + 101 * eps))) is False
    assert not zero_range(c((1, 1 + 101 * eps)))

    # Changed tol
    assert not zero_range(c((1, 1 + 2 * eps)), tol=eps)

    # Scaling up or down all the values has no effect since
    # the values are rescaled to 1 before checking against
    # the tolerance
    assert zero_range(100000 * c((1, 1 + eps)))
    assert not zero_range(100000 * c((1, 1 + 200 * eps)))
    assert zero_range(0.00001 * c((1, 1 + eps)))
    assert not zero_range(0.00001 * c((1, 1 + 200 * eps)))

    # NA values
    assert zero_range((1, np.nan))

    # Infinite values
    assert not zero_range((1, np.inf))
    assert not zero_range((-np.inf, np.inf))
    assert zero_range((np.inf, np.inf))

    # Single value
    assert zero_range(1)
    assert zero_range([1])

    # length greater than 2
    with pytest.raises(ValueError):
        zero_range([1, 2, 3])

    # datetime - pandas, cpython
    x = datetime(2010, 1, 1), datetime(2010, 1, 1)
    x2 = datetime(2010, 1, 1), datetime(2020, 1, 1)
    x3 = (
        pd.Timestamp("2010-01-01", tz="US/Eastern"),
        pd.Timestamp("2010-01-01", tz="US/Central"),
    )
    assert zero_range(x)
    assert not zero_range(x2)
    assert not zero_range(x3)

    # datetime - numpy
    x = np.datetime64(7, "D"), np.datetime64(7, "D")
    x2 = np.datetime64(7, "D"), np.datetime64(1, "W")
    x3 = np.datetime64(7, "D"), np.datetime64(1, "D")
    assert zero_range(x)
    assert zero_range(x2)
    assert not zero_range(x3)

    # timedelta - pandas, cpython
    x = timedelta(seconds=2010), timedelta(seconds=2010)
    x2 = (
        timedelta(seconds=2010, microseconds=90),
        timedelta(seconds=2010, microseconds=34),
    )
    x3 = pd.Timedelta(200, "D"), pd.Timedelta(203, "D")
    assert zero_range(x)
    assert not zero_range(x2)
    assert not zero_range(x3)

    # timedelta - numpy
    x = np.timedelta64(7, "D"), np.timedelta64(7, "D")
    x2 = np.timedelta64(7, "D"), np.timedelta64(1, "W")
    x3 = np.timedelta64(7, "D"), np.timedelta64(2, "D")
    assert zero_range(x)
    assert zero_range(x2)
    assert not zero_range(x3)

    # branches #
    assert str(zero_range([4, float("nan")])) == "nan"
    assert not zero_range([4, float("inf")])
    with pytest.raises(TypeError):
        zero_range(["a", "b"])

    # are handled without error
    x = np.array([8, 4], dtype=np.uint64)
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        zero_range(x)



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing import Literal, TypeAlias

    from mizani.colors.color_palette import palette

    RGB256Color: TypeAlias = tuple[int, int, int]
    RGB256Swatch: TypeAlias = list[RGB256Color]
    RGB256Swatches: TypeAlias = list[RGB256Swatch]

    RGBHexColor: TypeAlias = str
    RGBHexSwatch: TypeAlias = list[RGBHexColor]
    RGBHexSwatches: TypeAlias = list[RGBHexSwatch]

    ColorScheme: TypeAlias = Literal[
        "diverging",
        "qualitative",
        "sequential",
    ]
    ColorSchemeShort: TypeAlias = Literal[
        "div",
        "qual",
        "seq",
    ]
    ColorPalette: TypeAlias = palette



"""
Palettes are the link between data values and the values along
the dimension of a scale. Before a collection of values can be
represented on a scale, they are transformed by a palette. This
transformation is knowing as mapping. Values are mapped onto a
scale by a palette.

Scales tend to have restrictions on the magnitude of quantities
that they can intelligibly represent. For example, the size of
a point should be significantly smaller than the plot panel
onto which it is plotted or else it would be hard to compare
two or more points. Therefore palettes must be created that
enforce such restrictions. This is the reason for the ``*_pal``
functions that create and return the actual palette functions.
"""
from __future__ import annotations

import colorsys
import typing
from warnings import warn

import numpy as np
import pandas as pd

from .bounds import rescale
from .external import crayon_rgb, husl, xkcd_rgb
from .utils import identity

if typing.TYPE_CHECKING:
    from typing import Literal

    from mizani.colors.typing import ColorScheme, ColorSchemeShort


__all__ = [
    "hls_palette",
    "husl_palette",
    "rescale_pal",
    "area_pal",
    "abs_area",
    "grey_pal",
    "hue_pal",
    "brewer_pal",
    "gradient_n_pal",
    "cmap_pal",
    "cmap_d_pal",
    "desaturate_pal",
    "manual_pal",
    "xkcd_palette",
    "crayon_palette",
    "cubehelix_pal",
]


def hls_palette(n_colors=6, h=0.01, l=0.6, s=0.65):
    """
    Get a set of evenly spaced colors in HLS hue space.

    h, l, and s should be between 0 and 1

    Parameters
    ----------

    n_colors : int
        number of colors in the palette
    h : float
        first hue
    l : float
        lightness
    s : float
        saturation

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    See Also
    --------
    husl_palette : Make a palette using evenly spaced circular
        hues in the HUSL system.

    Examples
    --------
    >>> len(hls_palette(2))
    2
    >>> len(hls_palette(9))
    9
    """
    hues = np.linspace(0, 1, n_colors + 1)[:-1]
    hues += h
    hues %= 1
    hues -= hues.astype(int)
    palette = [colorsys.hls_to_rgb(h_i, l, s) for h_i in hues]
    return palette


def husl_palette(n_colors=6, h=0.01, s=0.9, l=0.65):
    """
    Get a set of evenly spaced colors in HUSL hue space.

    h, s, and l should be between 0 and 1

    Parameters
    ----------

    n_colors : int
        number of colors in the palette
    h : float
        first hue
    s : float
        saturation
    l : float
        lightness

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    See Also
    --------
    hls_palette : Make a palette using evenly spaced circular
        hues in the HSL system.

    Examples
    --------
    >>> len(husl_palette(3))
    3
    >>> len(husl_palette(11))
    11
    """
    hues = np.linspace(0, 1, n_colors + 1)[:-1]
    hues += h
    hues %= 1
    hues *= 359
    s *= 99
    l *= 99
    palette = [husl.husl_to_rgb(h_i, s, l) for h_i in hues]
    return palette


def rescale_pal(range=(0.1, 1)):
    """
    Rescale the input to the specific output range.

    Useful for alpha, size, and continuous position.

    Parameters
    ----------
    range : tuple
        Range of the scale

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in
        the specified range.

    Examples
    --------
    >>> palette = rescale_pal()
    >>> palette([0, .2, .4, .6, .8, 1])
    array([0.1 , 0.28, 0.46, 0.64, 0.82, 1.  ])

    The returned palette expects inputs in the ``[0, 1]``
    range. Any value outside those limits is clipped to
    ``range[0]`` or ``range[1]``.

    >>> palette([-2, -1, 0.2, .4, .8, 2, 3])
    array([0.1 , 0.1 , 0.28, 0.46, 0.82, 1.  , 1.  ])
    """

    def _rescale(x):
        return rescale(x, range, _from=(0, 1))

    return _rescale


def area_pal(range=(1, 6)):
    """
    Point area palette (continuous).

    Parameters
    ----------
    range : tuple
        Numeric vector of length two, giving range of possible sizes.
        Should be greater than 0.

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in
        the specified range.

    Examples
    --------
    >>> x = np.arange(0, .6, .1)**2
    >>> palette = area_pal()
    >>> palette(x)
    array([1. , 1.5, 2. , 2.5, 3. , 3.5])

    The results are equidistant because the input ``x`` is in
    area space, i.e it is squared.
    """

    def area_palette(x):
        return rescale(np.sqrt(x), to=range, _from=(0, 1))

    return area_palette


def abs_area(max):
    """
    Point area palette (continuous), with area proportional to value.

    Parameters
    ----------
    max : float
        A number representing the maximum size

    Returns
    -------
    out : function
        Palette function that takes a sequence of values
        in the range ``[0, 1]`` and returns values in the range
        ``[0, max]``.

    Examples
    --------
    >>> x = np.arange(0, .8, .1)**2
    >>> palette = abs_area(5)
    >>> palette(x)
    array([0. , 0.5, 1. , 1.5, 2. , 2.5, 3. , 3.5])

    Compared to :func:`area_pal`, :func:`abs_area` will handle values
    in the range ``[-1, 0]`` without returning ``np.nan``. And values
    whose absolute value is greater than 1 will be clipped to the
    maximum.
    """

    def abs_area_palette(x):
        return rescale(np.sqrt(np.abs(x)), to=(0, max), _from=(0, 1))

    return abs_area_palette


def grey_pal(start=0.2, end=0.8):
    """
    Utility for creating continuous grey scale palette

    Parameters
    ----------
    start : float
        grey value at low end of palette
    end : float
        grey value at high end of palette

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors.

    Examples
    --------
    >>> palette = grey_pal()
    >>> palette(5)
    ['#333333', '#737373', '#989898', '#b5b5b5', '#cccccc']
    """
    import matplotlib.colors as mcolors

    gamma = 2.2
    ends = ((0.0, start, start), (1.0, end, end))
    cdict = {"red": ends, "green": ends, "blue": ends}
    grey_cmap = mcolors.LinearSegmentedColormap("grey", cdict)

    def continuous_grey_palette(n):
        # The grey scale points are linearly separated in
        # gamma encoded space
        x = np.linspace(start**gamma, end**gamma, n)
        # Map points onto the [0, 1] palette domain
        vals = (x ** (1.0 / gamma) - start) / (end - start)
        return ratios_to_colors(vals, grey_cmap)

    return continuous_grey_palette


def hue_pal(h=0.01, l=0.6, s=0.65, color_space="hls"):
    """
    Utility for making hue palettes for color schemes.

    Parameters
    ----------
    h : float
        first hue. In the [0, 1] range
    l : float
        lightness. In the [0, 1] range
    s : float
        saturation. In the [0, 1] range
    color_space : 'hls' | 'husl'
        Color space to use for the palette

    Returns
    -------
    out : function
        A discrete color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors. Though the palette
        is continuous, since it is varies the hue it
        is good for categorical data. However if ``n``
        is large enough the colors show continuity.

    Examples
    --------
    >>> hue_pal()(5)
    ['#db5f57', '#b9db57', '#57db94', '#5784db', '#c957db']
    >>> hue_pal(color_space='husl')(5)
    ['#e0697e', '#9b9054', '#569d79', '#5b98ab', '#b675d7']
    """
    import matplotlib.colors as mcolors

    if not all(0 <= val <= 1 for val in (h, l, s)):
        msg = (
            "hue_pal expects values to be between 0 and 1. "
            " I got h={}, l={}, s={}".format(h, l, s)
        )
        raise ValueError(msg)

    if color_space not in ("hls", "husl"):
        msg = "color_space should be one of ['hls', 'husl']"
        raise ValueError(msg)

    name = "{}_palette".format(color_space)
    palette = globals()[name]

    def _hue_pal(n):
        colors = palette(n, h=h, l=l, s=s)
        return [mcolors.rgb2hex(c) for c in colors]

    return _hue_pal


def brewer_pal(
    type: ColorScheme | ColorSchemeShort = "seq",
    palette: int = 1,
    direction: Literal[1, -1] = 1,
):
    """
    Utility for making a brewer palette

    Parameters
    ----------
    type : 'sequential' | 'qualitative' | 'diverging'
        Type of palette. Sequential, Qualitative or
        Diverging. The following abbreviations may
        be used, ``seq``, ``qual`` or ``div``.
    palette : int | str
        Which palette to choose from. If is an integer,
        it must be in the range ``[0, m]``, where ``m``
        depends on the number sequential, qualitative or
        diverging palettes. If it is a string, then it
        is the name of the palette.
    direction : int
        The order of colours in the scale. If -1 the order
        of colors is reversed. The default is 1.

    Returns
    -------
    out : function
        A color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        colors. The maximum value of ``n`` varies
        depending on the parameters.

    Examples
    --------
    >>> brewer_pal()(5)
    ['#EFF3FF', '#BDD7E7', '#6BAED6', '#3182BD', '#08519C']
    >>> brewer_pal('qual')(5)
    ['#7FC97F', '#BEAED4', '#FDC086', '#FFFF99', '#386CB0']
    >>> brewer_pal('qual', 2)(5)
    ['#1B9E77', '#D95F02', '#7570B3', '#E7298A', '#66A61E']
    >>> brewer_pal('seq', 'PuBuGn')(5)
    ['#F6EFF7', '#BDC9E1', '#67A9CF', '#1C9099', '#016C59']

    The available color names for each palette type can be
    obtained using the following code::

        from mizani.colors.brewer import get_palette_names

        print(get_palette_names("sequential"))
        print(get_palette_names("qualitative"))
        print(get_palette_names("diverging"))
    """
    from .colors import brewer

    if direction != 1 and direction != -1:
        raise ValueError("direction should be 1 or -1.")

    pal = brewer.get_color_palette(type, palette)

    def _brewer_pal(n):
        # Only draw the maximum allowable colors from the palette
        # and fill any remaining spots with None
        _n = min(max(n, pal.min_colors), pal.max_colors)
        color_map = pal.get_hex_swatch(_n)
        colors = color_map[:n]
        if n > pal.max_colors:
            msg = (
                "Warning message:"
                f"Brewer palette {pal.name} has a maximum "
                f"of {pal.max_colors} colors Returning the palette you "
                "asked for with that many colors"
            )
            warn(msg)
            colors = colors + [None] * (n - pal.max_colors)
        return colors[::direction]

    return _brewer_pal


def ratios_to_colors(values, colormap):
    """
    Map values in the range [0, 1] onto colors

    Parameters
    ----------
    values : array_like | float
        Numeric(s) in the range [0, 1]
    colormap : cmap
        Matplotlib colormap to use for the mapping

    Returns
    -------
    out : list | float
        Color(s) corresponding to the values
    """
    import matplotlib.colors as mcolors

    iterable = True
    try:
        iter(values)
    except TypeError:
        iterable = False
        values = [values]

    color_tuples = colormap(values)
    hex_colors = [mcolors.rgb2hex(t) for t in color_tuples]

    nan_bool_idx = pd.isnull(values) | np.isinf(values)
    if any(nan_bool_idx):
        hex_colors = [
            np.nan if is_nan else color
            for color, is_nan in zip(hex_colors, nan_bool_idx)
        ]
    return hex_colors if iterable else hex_colors[0]


def gradient_n_pal(colors, values=None, name="gradientn"):
    """
    Create a n color gradient palette

    Parameters
    ----------
    colors : list
        list of colors
    values : list, optional
        list of points in the range [0, 1] at which to
        place each color. Must be the same size as
        `colors`. Default to evenly space the colors
    name : str
        Name to call the resultant MPL colormap

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = gradient_n_pal(['red', 'blue'])
    >>> palette([0, .25, .5, .75, 1])
    ['#ff0000', '#bf0040', '#7f0080', '#3f00c0', '#0000ff']
    >>> palette([-np.inf, 0, np.nan, 1, np.inf])
    [nan, '#ff0000', nan, '#0000ff', nan]
    """
    import matplotlib.colors as mcolors

    # Note: For better results across devices and media types,
    # it would be better to do the interpolation in
    # Lab color space.
    if values is None:
        colormap = mcolors.LinearSegmentedColormap.from_list(name, colors)
    else:
        colormap = mcolors.LinearSegmentedColormap.from_list(
            name, list(zip(values, colors))
        )

    def _gradient_n_pal(vals):
        return ratios_to_colors(vals, colormap)

    return _gradient_n_pal


def cmap_pal(name, lut=None):
    """
    Create a continuous palette using an MPL colormap

    Parameters
    ----------
    name : str
        Name of colormap
    lut : None | int
        This is the number of entries desired in the lookup table.
        Default is ``None``, leave it up Matplotlib.

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = cmap_pal('viridis')
    >>> palette([.1, .2, .3, .4, .5])
    ['#482475', '#414487', '#355f8d', '#2a788e', '#21918c']
    """
    import matplotlib as mpl

    colormap = mpl.colormaps[name]

    if lut is not None:
        warn(
            "The lut parameter has been deprecated and will "
            "be removed in a future version.",
            FutureWarning,
        )
        resample = getattr(
            colormap, "resampled", getattr(colormap, "_resample", None)
        )
        colormap = resample(lut)

    def _cmap_pal(vals):
        return ratios_to_colors(vals, colormap)

    return _cmap_pal


def cmap_d_pal(name, lut=None):
    """
    Create a discrete palette using an MPL Listed colormap

    Parameters
    ----------
    name : str
        Name of colormap
    lut : None | int
        This is the number of entries desired in the lookup table.
        Default is ``None``, leave it up Matplotlib.

    Returns
    -------
    out : function
        A discrete color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        colors. The maximum value of ``n`` varies
        depending on the parameters.

    Examples
    --------
    >>> palette = cmap_d_pal('viridis')
    >>> palette(5)
    ['#440154', '#3b528b', '#21918c', '#5cc863', '#fde725']
    """
    import matplotlib as mpl
    import matplotlib.colors as mcolors

    colormap = mpl.colormaps[name]

    if lut is not None:
        warn(
            "The lut parameter has been deprecated and will "
            "be removed in a future version.",
            FutureWarning,
        )
        resample = getattr(
            colormap, "resampled", getattr(colormap, "_resample", None)
        )
        colormap = resample(lut)

    if not isinstance(colormap, mcolors.ListedColormap):
        raise ValueError(
            "For a discrete palette, cmap must be of type "
            "matplotlib.colors.ListedColormap"
        )

    ncolors = len(colormap.colors)

    def _cmap_d_pal(n):
        if n > ncolors:
            raise ValueError(
                "cmap `{}` has {} colors you requested {} "
                "colors.".format(name, ncolors, n)
            )

        if ncolors < 256:
            return [mcolors.rgb2hex(c) for c in colormap.colors[:n]]
        else:
            # Assume these are continuous and get colors equally spaced
            # intervals  e.g. viridis is defined with 256 colors
            idx = np.linspace(0, ncolors - 1, n).round().astype(int)
            return [mcolors.rgb2hex(colormap.colors[i]) for i in idx]

    return _cmap_d_pal


def desaturate_pal(color, prop, reverse=False):
    """
    Create a palette that desaturate a color by some proportion

    Parameters
    ----------
    color : matplotlib color
        hex, rgb-tuple, or html color name
    prop : float
        saturation channel of color will be multiplied by
        this value
    reverse : bool
        Whether to reverse the palette.

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        parameter either a :class:`float` or a sequence
        of floats maps those value(s) onto the palette
        and returns color(s). The float(s) must be
        in the range [0, 1].

    Examples
    --------
    >>> palette = desaturate_pal('red', .1)
    >>> palette([0, .25, .5, .75, 1])
    ['#ff0000', '#e21d1d', '#c53a3a', '#a95656', '#8c7373']
    """
    import matplotlib.colors as mcolors

    if not 0 <= prop <= 1:
        raise ValueError("prop must be between 0 and 1")

    # Get rgb tuple rep
    # Convert to hls
    # Desaturate the saturation channel
    # Convert back to rgb
    rgb = mcolors.colorConverter.to_rgb(color)
    h, l, s = colorsys.rgb_to_hls(*rgb)
    s *= prop
    desaturated_color = colorsys.hls_to_rgb(h, l, s)
    colors = [color, desaturated_color]
    if reverse:
        colors = colors[::-1]
    return gradient_n_pal(colors, name="desaturated")


def manual_pal(values):
    """
    Create a palette from a list of values

    Parameters
    ----------
    values : sequence
        Values that will be returned by the palette function.

    Returns
    -------
    out : function
        A function palette that takes a single
        :class:`int` parameter ``n`` and returns ``n`` values.

    Examples
    --------
    >>> palette = manual_pal(['a', 'b', 'c', 'd', 'e'])
    >>> palette(3)
    ['a', 'b', 'c']
    """
    max_n = len(values)

    def _manual_pal(n):
        if n > max_n:
            msg = (
                "Palette can return a maximum of {} values. "
                "{} values requested."
            )
            warn(msg.format(max_n, n))

        return values[:n]

    return _manual_pal


def xkcd_palette(colors):
    """
    Make a palette with color names from the xkcd color survey.

    See xkcd for the full list of colors: http://xkcd.com/color/rgb/

    Parameters
    ----------
    colors : list of strings
        List of keys in the ``mizani.external.xkcd_rgb`` dictionary.

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    Examples
    --------
    >>> palette = xkcd_palette(['red', 'green', 'blue'])
    >>> palette
    ['#e50000', '#15b01a', '#0343df']

    >>> from mizani.external import xkcd_rgb
    >>> list(sorted(xkcd_rgb.keys()))[:5]
    ['acid green', 'adobe', 'algae', 'algae green', 'almost black']
    """
    return [xkcd_rgb[name] for name in colors]


def crayon_palette(colors):
    """
    Make a palette with color names from Crayola crayons.

    The colors come from
    http://en.wikipedia.org/wiki/List_of_Crayola_crayon_colors

    Parameters
    ----------
    colors : list of strings
        List of keys in the ``mizani.external.crayloax_rgb`` dictionary.

    Returns
    -------
    palette : list
        List of colors as RGB hex strings.

    Examples
    --------
    >>> palette = crayon_palette(['almond', 'silver', 'yellow'])
    >>> palette
    ['#eed9c4', '#c9c0bb', '#fbe870']

    >>> from mizani.external import crayon_rgb
    >>> list(sorted(crayon_rgb.keys()))[:5]
    ['almond', 'antique brass', 'apricot', 'aquamarine', 'asparagus']
    """
    return [crayon_rgb[name] for name in colors]


def cubehelix_pal(
    start=0, rot=0.4, gamma=1.0, hue=0.8, light=0.85, dark=0.15, reverse=False
):
    """
    Utility for creating continuous palette from the cubehelix system.

    This produces a colormap with linearly-decreasing (or increasing)
    brightness. That means that information will be preserved if printed to
    black and white or viewed by someone who is colorblind.

    Parameters
    ----------
    start : float (0 <= start <= 3)
        The hue at the start of the helix.
    rot : float
        Rotations around the hue wheel over the range of the palette.
    gamma : float (0 <= gamma)
        Gamma factor to emphasize darker (gamma < 1) or lighter (gamma > 1)
        colors.
    hue : float (0 <= hue <= 1)
        Saturation of the colors.
    dark : float (0 <= dark <= 1)
        Intensity of the darkest color in the palette.
    light : float (0 <= light <= 1)
        Intensity of the lightest color in the palette.
    reverse : bool
        If True, the palette will go from dark to light.

    Returns
    -------
    out : function
        Continuous color palette that takes a single
        :class:`int` parameter ``n`` and returns ``n``
        equally spaced colors.


    References
    ----------
    Green, D. A. (2011). "A colour scheme for the display of astronomical
    intensity images". Bulletin of the Astromical Society of India, Vol. 39,
    p. 289-295.

    Examples
    --------
    >>> palette = cubehelix_pal()
    >>> palette(5)
    ['#edd1cb', '#d499a7', '#aa688f', '#6e4071', '#2d1e3e']
    """
    import matplotlib as mpl
    import matplotlib.colors as mcolors

    cdict = mpl._cm.cubehelix(gamma, start, rot, hue)
    cubehelix_cmap = mcolors.LinearSegmentedColormap("cubehelix", cdict)

    def cubehelix_palette(n):
        values = np.linspace(light, dark, n)
        return [mcolors.rgb2hex(cubehelix_cmap(x)) for x in values]

    return cubehelix_palette


def identity_pal():
    """
    Create palette that maps values onto themselves

    Returns
    -------
    out : function
        Palette function that takes a value or sequence of values
        and returns the same values.

    Examples
    --------
    >>> palette = identity_pal()
    >>> palette(9)
    9
    >>> palette([2, 4, 6])
    [2, 4, 6]
    """
    return identity



"""
Standard colors found in regular Crayola assortments from 1903 to
the present

Source: https://en.wikipedia.org/wiki/List_of_Crayola_crayon_colors
"""

crayon_rgb = {
    'almond': '#eed9c4',
    'antique brass': '#c88a65',
    'apricot': '#fdd5b1',
    'aquamarine': '#95e0e8',
    'asparagus': '#7ba05b',
    'banana mania': '#fbe7b2',
    'beaver': '#926f5b',
    'bittersweet': '#fe6f5e',
    'black': '#000000',
    'blue (i)': '#4997d0',
    'blue (ii)': '#4570e6',
    'blue (iii)': '#0066ff',
    'blue bell': '#9999cc',
    'blue-gray': '#c8c8cd',
    'blue-green': '#0095b7',
    'blue-violet': '#6456b7',
    'blush': '#db5079',
    'brick red': '#c62d42',
    'brilliant rose': '#e667ce',
    'brown': '#af593e',
    'burnt orange': '#ff7f49',
    'burnt sienna': '#e97451',
    'burnt umber': '#805533',
    'cadet blue': '#a9b2c3',
    'canary': '#ffff99',
    'caribbean green': '#00cc99',
    'carmine': '#e62e6b',
    'carnation pink': '#ffa6c9',
    'celestial blue': '#7070cc',
    'cerise': '#da3287',
    'cerulean': '#02a4d3',
    'cerulean blue': '#339acc',
    'charcoal gray': '#736a62',
    'cobalt blue': '#8c90c8',
    'copper': '#da8a67',
    'cornflower': '#93ccea',
    'cotton candy': '#ffb7d5',
    'dandelion': '#fed85d',
    'dark venetian red': '#b33b24',
    'denim': '#1560bd',
    'desert sand': '#edc9af',
    'eggplant': '#614051',
    'english vermilion': '#cc474b',
    'fern': '#63b76c',
    'forest green': '#5fa777',
    'fuchsia': '#c154c1',
    'fuzzy wuzzy': '#87421f',
    'gold (i)': '#92926e',
    'gold (ii)': '#e6be8a',
    'goldenrod': '#fcd667',
    'granny smith apple': '#9de093',
    'gray': '#8b8680',
    'green': '#3aa655',
    'green-blue': '#2887c8',
    'green-yellow': '#f1e788',
    'inchworm': '#afe313',
    'indian red': '#b94e48',
    'indigo': '#4f69c6',
    'jazzberry jam': '#a50b5e',
    'jungle green': '#29ab87',
    'lavender (i)': '#bf8fcc',
    'lavender (ii)': '#fbaed2',
    'lemon yellow': '#ffff9f',
    'light blue': '#8fd8d8',
    'light chrome green': '#bee64b',
    'light venetian red': '#e6735c',
    'macaroni and cheese': '#ffb97b',
    'madder lake': '#cc3336',
    'magenta': '#f653a6',
    'mahogany': '#ca3435',
    'maize': '#f2c649',
    'manatee': '#8d90a1',
    'mango tango': '#e77200',
    'maroon': '#c32148',
    'mauvelous': '#f091a9',
    'maximum blue': '#47abcc',
    'maximum blue green': '#30bfbf',
    'maximum blue purple': '#acace6',
    'maximum green': '#5e8c31',
    'maximum green yellow': '#d9e650',
    'maximum purple': '#733380',
    'maximum red': '#d92121',
    'maximum red purple': '#a63a79',
    'maximum yellow': '#fafa37',
    'maximum yellow red': '#f2ba49',
    'medium chrome green': '#6ca67c',
    'medium rose': '#d96cbe',
    'medium violet': '#8f47b3',
    'melon': '#febaad',
    'middle blue': '#7ed4e6',
    'middle blue green': '#8dd9cc',
    'middle blue purple': '#8b72be',
    'middle green': '#4d8c57',
    'middle green yellow': '#acbf60',
    'middle purple': '#d982b5',
    'middle red': '#e58e73',
    'middle red purple': '#a55353',
    'middle yellow': '#ffeb00',
    'middle yellow red': '#ecb176',
    'midnight blue': '#00468c',
    'mountain meadow': '#1ab385',
    'mulberry': '#c8509b',
    'navy blue': '#0066cc',
    'olive green': '#b5b35c',
    'orange': '#ff8833',
    'orange-red': '#ff5349',
    'orange-yellow': '#f8d568',
    'orchid': '#e29cd2',
    'outer space': '#2d383a',
    'pacific blue': '#009dc4',
    'peach': '#ffcba4',
    'periwinkle': '#c3cde6',
    'permanent geranium lake': '#e12c2c',
    'pig pink': '#fdd7e4',
    'pine green': '#01786f',
    'pink flamingo': '#fc74fd',
    'pink sherbert': '#f7a38e',
    'plum': '#8e3179',
    'purple heart': '#652dc1',
    "purple mountains' majesty": '#d6aedd',
    'raw sienna': '#d27d46',
    'raw sienna (i)': '#e6bc5c',
    'raw umber': '#665233',
    'razzmatazz': '#e30b5c',
    'red': '#ed0a3f',
    'red-orange': '#ff681f',
    'red-violet': '#bb3385',
    "robin's egg blue": '#00cccc',
    'royal purple': '#6b3fa0',
    'salmon': '#ff91a4',
    'scarlet': '#fd0e35',
    'sea green': '#93dfb8',
    'sepia': '#9e5b40',
    'shadow': '#837050',
    'shamrock': '#33cc99',
    'silver': '#c9c0bb',
    'sky blue': '#76d7ea',
    'spring green': '#ecebbd',
    'sunset orange': '#fe4c40',
    'tan': '#d99a6c',
    'teal blue': '#008080',
    'thistle': '#ebb0d7',
    'tickle me pink': '#fc80a5',
    'timberwolf': '#d9d6cf',
    'tropical rain forest': '#00755e',
    'tumbleweed': '#dea681',
    'turquoise blue': '#6cdae7',
    'ultramarine blue': '#3f26bf',
    'van dyke brown': '#664228',
    'venetian red': '#cc553d',
    'violet (i)': '#732e6c',
    'violet (ii)': '#8359a3',
    'violet-blue': '#766ec8',
    'violet-red': '#f7468a',
    'vivid tangerine': '#ff9980',
    'vivid violet': '#803790',
    'white': '#ffffff',
    'wild blue yonder': '#7a89b8',
    'wild strawberry': '#ff3399',
    'wisteria': '#c9a0dc',
    'yellow': '#fbe870',
    'yellow-green': '#c5e17a',
    'yellow-orange': '#ffae42',
}



from __future__ import annotations

import typing
from collections import OrderedDict, defaultdict
from collections.abc import Iterator
from datetime import timezone, tzinfo
from itertools import chain
from warnings import warn

import numpy as np

if typing.TYPE_CHECKING:
    from datetime import datetime
    from typing import Sequence


__all__ = [
    "round_any",
    "min_max",
    "match",
    "precision",
    "first_element",
    "multitype_sort",
    "same_log10_order_of_magnitude",
    "identity",
    "get_categories",
    "get_timezone",
]

DISCRETE_KINDS = "ObUS"
CONTINUOUS_KINDS = "ifuc"

SECONDS = OrderedDict(
    [
        ("ns", 1e-9),  # nanosecond
        ("us", 1e-6),  # microsecond
        ("ms", 1e-3),  # millisecond
        ("s", 1),  # second
        ("m", 60),  # month
        ("h", 3600),  # hour
        ("d", 24 * 3600),  # day
        ("w", 7 * 24 * 3600),  # week
        ("M", 31 * 24 * 3600),  # month
        ("y", 365 * 24 * 3600),  # year
    ]
)

NANOSECONDS = OrderedDict(
    [
        ("ns", 1),  # nanosecond
        ("us", 1e3),  # microsecond
        ("ms", 1e6),  # millisecond
        ("s", 1e9),  # second
        ("m", 60e9),  # month
        ("h", 3600e9),  # hour
        ("d", 24 * 3600e9),  # day
        ("w", 7 * 24 * 3600e9),  # week
        ("M", 31 * 24 * 3600e9),  # month
        ("y", 365 * 24 * 3600e9),  # year
    ]
)


def round_any(x, accuracy, f=np.round):
    """
    Round to multiple of any number.
    """
    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    return f(x / accuracy) * accuracy


def min_max(x, na_rm=False, finite=True):
    """
    Return the minimum and maximum of x

    Parameters
    ----------
    x : array_like
        Sequence
    na_rm : bool
        Whether to remove ``nan`` values.
    finite : bool
        Whether to consider only finite values.

    Returns
    -------
    out : tuple
        (minimum, maximum) of x
    """
    if not hasattr(x, "dtype"):
        x = np.asarray(x)

    if na_rm and finite:
        x = x[np.isfinite(x)]
    elif not na_rm and np.any(np.isnan(x)):
        return np.nan, np.nan
    elif na_rm:
        x = x[~np.isnan(x)]
    elif finite:
        x = x[~np.isinf(x)]

    if len(x):
        return np.min(x), np.max(x)
    else:
        return float("-inf"), float("inf")


def match(v1, v2, nomatch=-1, incomparables=None, start=0):
    """
    Return a vector of the positions of (first)
    matches of its first argument in its second.

    Parameters
    ----------
    v1: array_like
        Values to be matched

    v2: array_like
        Values to be matched against

    nomatch: int
        Value to be returned in the case when
        no match is found.

    incomparables: array_like
        Values that cannot be matched. Any value in ``v1``
        matching a value in this list is assigned the nomatch
        value.
    start: int
        Type of indexing to use. Most likely 0 or 1
    """
    v2_indices = {}
    for i, x in enumerate(v2):
        if x not in v2_indices:
            v2_indices[x] = i

    v1_to_v2_map = [nomatch] * len(v1)
    skip = set(incomparables) if incomparables else set()
    for i, x in enumerate(v1):
        if x in skip:
            continue

        try:
            v1_to_v2_map[i] = v2_indices[x] + start
        except KeyError:
            pass

    return v1_to_v2_map


def precision(x):
    """
    Return the precision of x

    Parameters
    ----------
    x : array_like | numeric
        Value(s) whose for which to compute the precision.

    Returns
    -------
    out : numeric
        The precision of ``x`` or that the values in ``x``.

    Notes
    -----
    The precision is computed in base 10.

    Examples
    --------
    >>> precision(0.08)
    0.01
    >>> precision(9)
    1
    >>> precision(16)
    10
    """
    from .bounds import zero_range

    rng = min_max(x, na_rm=True)
    if zero_range(rng):
        span = np.abs(rng[0])
    else:
        span = np.diff(rng)[0]

    if span == 0:
        return 1
    else:
        return 10 ** int(np.floor(np.log10(span)))


def first_element(obj):
    """
    Return the first element of `obj`

    Parameters
    ----------
    obj : iterable
        Should not be an iterator

    Returns
    -------
    out : object
        First element of `obj`. Raise a class:`StopIteration`
        exception if `obj` is empty.
    """
    if isinstance(obj, Iterator):
        raise RuntimeError("Cannot get the first element of an iterator")
    return next(iter(obj))


def multitype_sort(a):
    """
    Sort elements of multiple types

    x is assumed to contain elements of different types, such that
    plain sort would raise a `TypeError`.

    Parameters
    ----------
    a : array-like
        Array of items to be sorted

    Returns
    -------
    out : list
        Items sorted within their type groups.
    """
    types = defaultdict(list)
    numbers = {int, float, complex}

    for x in a:
        t = type(x)
        if t in numbers:
            types["number"].append(x)
        else:
            types[t].append(x)

    for t in types:
        types[t] = np.sort(types[t])

    return list(chain.from_iterable(types[t] for t in types))


def same_log10_order_of_magnitude(x, delta=0.1):
    """
    Return true if range is approximately in same order of magnitude

    For example these sequences are in the same order of magnitude:

        - [1, 8, 5]     # [1, 10)
        - [35, 20, 80]  # [10 100)
        - [232, 730]    # [100, 1000)

    Parameters
    ----------
    x : array-like
         Values in base 10. Must be size 2 and
        ``rng[0] <= rng[1]``.
    delta : float
        Fuzz factor for approximation. It is multiplicative.
    """
    dmin = np.log10(np.min(x) * (1 - delta))
    dmax = np.log10(np.max(x) * (1 + delta))
    return np.floor(dmin) == np.floor(dmax)


def identity(*args):
    """
    Return whatever is passed in
    """
    return args if len(args) > 1 else args[0]


def get_categories(x):
    """
    Return the categories of x

    Parameters
    ----------
    x : category_like
        Input Values

    Returns
    -------
    out : Index
        Categories of x
    """
    try:
        return x.cat.categories  # series
    except AttributeError:
        try:
            return x.categories  # plain categorical
        except AttributeError:
            raise TypeError("x is the wrong type, it has no categories")


def log(x, base):
    """
    Calculate the log of x

    Parameters
    ----------
    x : category_like
        Input Values

    base : float
        Base of logarithm

    Returns
    -------
    out : float
        Log of x
    """
    if base == 10:
        res = np.log10(x)
    elif base == 2:
        res = np.log2(x)
    elif base == np.e:
        res = np.log(x)
    else:
        res = np.log(x) / np.log(base)
    return res


def get_timezone(x: Sequence[datetime]) -> tzinfo:
    """
    Return a single timezone for the sequence of datetimes

    Returns the timezone of first item and warns if any other item
    has a different timezone
    """

    # Ref: https://en.wikipedia.org/wiki/List_of_tz_database_time_zones

    info = x[0].tzinfo
    if info is None:
        return timezone.utc

    # Consistency check
    tzname0 = info.tzname(x[0])
    tznames = (dt.tzinfo.tzname(dt) if dt.tzinfo else None for dt in x)
    if any(tzname0 != name for name in tznames):
        msg = (
            "Dates in column have different time zones. "
            "Choosen `{}` the time zone of the first date. "
            "To use a different time zone, create a "
            "formatter and pass the time zone."
        )
        warn(msg.format(tzname0))
    return info
